# Copyright © The Debusine Developers
# See the AUTHORS file at the top-level directory of this distribution
#
# This file is part of Debusine. It is subject to the license terms
# in the LICENSE file found in the top-level directory of this
# distribution. No part of Debusine, including this file, may be copied,
# modified, propagated, or distributed except according to the terms
# contained in the LICENSE file.

"""Test views for package archives."""

import base64
import datetime as dt
import hashlib
from abc import ABC
from collections.abc import Mapping
from contextlib import ExitStack
from operator import attrgetter
from pathlib import PurePath
from typing import Any, ClassVar, overload, override
from unittest import mock

from debian.copyright import parse_multiline
from debian.deb822 import Deb822
from django.conf import settings
from django.test import override_settings
from django.utils import timezone
from django.utils.cache import has_vary_header
from pgpy import PGPKey, PGPKeyring
from pgpy.constants import EllipticCurveOID, PubKeyAlgorithm
from rest_framework import HTTP_HEADER_ENCODING, status

from debusine.artifacts.models import CollectionCategory
from debusine.assets.models import KeyPurpose
from debusine.db.models import Collection, CollectionRelation, Token, Workspace
from debusine.db.playground import scenarios
from debusine.test.django import TestCase, TestResponseType
from debusine.web.archives.views import RenderableSuite, SigningKeysResponse


class SigningKeysResponseTests(TestCase):
    """Tests for :py:class:`SigningKeysResponse`."""

    def test_render(self) -> None:
        keys = [
            PGPKey.new(PubKeyAlgorithm.EdDSA, EllipticCurveOID.Ed25519).pubkey
            for _ in range(2)
        ]
        response = SigningKeysResponse([bytes(key) for key in keys])
        self.assertEqual(response.status_code, status.HTTP_200_OK)
        self.assertEqual(
            response.headers["Content-Type"],
            "application/pgp-keys; charset=utf-8",
        )
        assert response.content.startswith(
            b"-----BEGIN PGP PUBLIC KEY BLOCK-----\n"
        )
        self.assertEqual(
            PGPKeyring(response.content).fingerprints(),
            {key.fingerprint for key in keys},
        )


class ArchiveViewTests(TestCase, ABC):
    """Helper methods for testing archive views."""

    playground_memory_file_store = False
    scenario = scenarios.DefaultContext()

    archive: ClassVar[Collection]
    suites: ClassVar[dict[str, Collection]]

    # Path used for some tests of generic behaviour in each subclass.
    example_path: str
    # Is the example path scoped to a suite?
    expect_in_suite: bool
    # Are paths tested here normally mutable?
    expect_mutable_no_archive: bool
    expect_mutable: bool
    # What Content-Type does this endpoint return on success?
    expect_content_type: str = "text/html; charset=utf-8"

    @override
    @classmethod
    def setUpTestData(cls) -> None:
        super().setUpTestData()
        cls.archive = cls.scenario.workspace.get_singleton_collection(
            pc=cls.scenario.context_ui(), category=CollectionCategory.ARCHIVE
        )
        cls.suites = {
            name: cls.playground.create_collection(
                name, CollectionCategory.SUITE, workspace=cls.scenario.workspace
            )
            for name in ("bookworm", "trixie")
        }

    @overload
    def format_snapshot(self, snapshot: None) -> None: ...

    @overload
    def format_snapshot(self, snapshot: dt.datetime) -> str: ...

    def format_snapshot(self, snapshot: dt.datetime | None) -> str | None:
        """Format a snapshot timestamp as a URL segment."""
        return None if snapshot is None else snapshot.strftime("%Y%m%dT%H%M%SZ")

    def archive_file_get(
        self,
        *,
        workspace_name: str | None = None,
        snapshot: dt.datetime | None = None,
        path: str,
        headers: Mapping[str, Any] | None = None,
        token: Token | None = None,
    ) -> TestResponseType:
        """Get a file from an archive."""
        with ExitStack() as stack:
            fqdn = settings.DEBUSINE_DEBIAN_ARCHIVE_PRIMARY_FQDN
            scope_name = self.scenario.scope.name
            if workspace_name is None:
                workspace_name = self.scenario.workspace.name
            url = f"/{scope_name}/{workspace_name}"
            if snapshot is not None:
                url += "/" + self.format_snapshot(snapshot)
            url += f"/{path}"
            if token is not None:
                headers = {**(headers or {})}
                if token.user is not None:
                    token_username = token.user.username
                else:
                    assert token.external_username
                    token_username = token.external_username
                credentials = f"{token_username}:{token.key}"
                base64_credentials = base64.b64encode(
                    credentials.encode(HTTP_HEADER_ENCODING)
                ).decode("ascii")
                headers["Authorization"] = "Basic " + base64_credentials
                stack.enter_context(
                    mock.patch(
                        "debusine.server.middlewares.scopes."
                        "AuthorizationMiddleware._basic_auth_allowed",
                        return_value=True,
                    )
                )
            return self.client.get(url, HTTP_HOST=fqdn, headers=headers)

    def assert_has_caching_headers(
        self, response: TestResponseType, *, mutable: bool
    ) -> None:
        self.assertEqual(
            response.headers["Cache-Control"],
            "max-age=1800, proxy-revalidate" if mutable else "max-age=31536000",
        )
        self.assertTrue(has_vary_header(response, "Authorization"))

    def assertResponse404(
        self, response: TestResponseType, message: str
    ) -> None:
        """Check that the response is a 404 with the given message."""
        self.assertEqual(response.status_code, status.HTTP_404_NOT_FOUND)
        self.assertEqual(response.context["exception"], message)

    def test_workspace_not_found(self) -> None:
        for snapshot in (None, timezone.now()):
            with self.subTest(snapshot=self.format_snapshot(snapshot)):
                response = self.archive_file_get(
                    workspace_name="nonexistent",
                    snapshot=snapshot,
                    path=self.example_path,
                )

                self.assertResponse404(
                    response,
                    "Workspace nonexistent not found in scope debusine,"
                    " or you are not authorized to see it",
                )
                self.assert_has_caching_headers(
                    response,
                    mutable=self.expect_mutable_no_archive and snapshot is None,
                )

    def test_archive_not_found(self) -> None:
        self.archive.child_items.all().delete()
        self.archive.delete()

        for snapshot in (None, timezone.now()):
            with self.subTest(snapshot=self.format_snapshot(snapshot)):
                response = self.archive_file_get(
                    snapshot=snapshot, path=self.example_path
                )

                self.assertResponse404(
                    response,
                    f"No {CollectionCategory.ARCHIVE} collection found in "
                    f"{self.scenario.workspace}",
                )
                self.assert_has_caching_headers(
                    response,
                    mutable=self.expect_mutable_no_archive and snapshot is None,
                )

    def test_private_workspace_not_accessible(self) -> None:
        self.scenario.workspace.public = False
        self.scenario.workspace.save()
        response = self.archive_file_get(path=self.example_path)
        self.assertEqual(response.status_code, status.HTTP_404_NOT_FOUND)

    def test_private_workspace_logged_in(self) -> None:
        self.scenario.workspace.public = False
        self.scenario.workspace.save()
        token = self.scenario.create_user_token()
        with self.scenario.assign_role(
            self.scenario.workspace, Workspace.Roles.CONTRIBUTOR
        ):
            response = self.archive_file_get(
                path=self.example_path, token=token
            )
        self.assertEqual(response.status_code, status.HTTP_200_OK)
        self.assertEqual(
            response.headers["Content-Type"], self.expect_content_type
        )
        if self.expect_content_type.startswith("text/html"):
            self.assertHTMLValid(response.content)

    def test_private_workspace_external_user_token_grants_archive(self) -> None:
        self.scenario.workspace.public = False
        self.scenario.workspace.save()
        token = self.playground.create_external_user_token(username="test")
        self.archive.grant_role(
            token, Collection.Roles.VIEWER, created_by=self.scenario.user
        )
        response = self.archive_file_get(path=self.example_path, token=token)
        self.assertEqual(response.status_code, status.HTTP_200_OK)
        self.assertEqual(
            response.headers["Content-Type"], self.expect_content_type
        )
        if self.expect_content_type.startswith("text/html"):
            self.assertHTMLValid(response.content)

    def test_private_workspace_external_user_token_grants_suite(self) -> None:
        self.scenario.workspace.public = False
        self.scenario.workspace.save()
        token = self.playground.create_external_user_token(username="test")
        self.suites["bookworm"].grant_role(
            token, Collection.Roles.VIEWER, created_by=self.scenario.user
        )
        response = self.archive_file_get(path=self.example_path, token=token)
        if self.expect_in_suite:
            self.assertEqual(response.status_code, status.HTTP_200_OK)
            self.assertEqual(
                response.headers["Content-Type"], self.expect_content_type
            )
            if self.expect_content_type.startswith("text/html"):
                self.assertHTMLValid(response.content)
        else:
            self.assertResponse404(
                response,
                f"Workspace {self.scenario.workspace.name} not found in scope"
                f" {self.scenario.scope}, or you are not authorized to see it",
            )

    def test_private_workspace_external_user_token_grants_wrong_suite(
        self,
    ) -> None:
        self.scenario.workspace.public = False
        self.scenario.workspace.save()
        token = self.playground.create_external_user_token(username="test")
        self.suites["trixie"].grant_role(
            token, Collection.Roles.VIEWER, created_by=self.scenario.user
        )
        response = self.archive_file_get(path=self.example_path, token=token)
        self.assertResponse404(
            response,
            f"Workspace {self.scenario.workspace.name} not found in scope"
            f" {self.scenario.scope}, or you are not authorized to see it",
        )


class ArchiveFileViewTests(ArchiveViewTests, ABC):
    """Helper methods for testing archive file views."""

    # A path that exercises the view under test, but that is not expected to
    # exist.
    example_path_nonexistent: str

    def assert_response_ok(
        self, response: TestResponseType, *, contents: bytes, mutable: bool
    ) -> None:
        self.assertEqual(response.status_code, status.HTTP_200_OK)
        self.assertEqual(
            b"".join(getattr(response, "streaming_content")), contents
        )
        self.assert_has_caching_headers(response, mutable=mutable)

    def test_file_not_found(self) -> None:
        for snapshot in (None, timezone.now()):
            with self.subTest(snapshot=self.format_snapshot(snapshot)):
                response = self.archive_file_get(
                    snapshot=snapshot, path=self.example_path_nonexistent
                )

                self.assertResponse404(
                    response, "No FileInArtifact matches the given query."
                )
                self.assert_has_caching_headers(
                    response, mutable=self.expect_mutable and snapshot is None
                )


class ArchiveRootViewTests(ArchiveViewTests):
    """Test retrieving the root of an archive."""

    example_path = ""
    expect_in_suite = False
    expect_mutable_no_archive = True
    expect_mutable = True

    def test_current(self) -> None:
        self.playground.create_collection(
            "not-exported",
            CollectionCategory.SUITE,
            workspace=self.scenario.workspace,
            data={"exported": False},
        )

        response = self.archive_file_get(path="")

        tree = self.assertResponseHTML(response)
        archive_suites = tree.xpath("//div[@id='archive-suites']")[0]
        self.assertTextContentEqual(
            archive_suites.p, "This archive contains the following suites:"
        )
        self.assertEqual(len(archive_suites.ul.li), len(self.suites))
        fqdn = settings.DEBUSINE_DEBIAN_ARCHIVE_FQDN
        workspace = self.scenario.workspace
        for li, suite in zip(
            archive_suites.ul.li,
            sorted(self.suites.values(), key=attrgetter("name")),
        ):
            self.assertTextContentEqual(li.p, f"{suite.name}: ")
            self.assertEqual(
                Deb822("".join(li.pre.itertext())),
                {
                    "Types": "deb",
                    "URIs": (
                        f"http://{fqdn}/{workspace.scope.name}/{workspace.name}"
                    ),
                    "Suites": suite.name,
                },
            )
        self.assert_has_caching_headers(response, mutable=True)

    def test_current_signing_keys(self) -> None:
        keys = [
            PGPKey.new(PubKeyAlgorithm.EdDSA, EllipticCurveOID.Ed25519).pubkey
            for _ in range(len(self.suites))
        ]
        sorted_suites = sorted(self.suites.values(), key=attrgetter("name"))
        for suite, key in zip(sorted_suites, keys):
            self.playground.create_asset_usage(
                resource=self.playground.create_signing_key_asset(
                    purpose=KeyPurpose.OPENPGP_REPOSITORY,
                    fingerprint=key.fingerprint,
                    public_key=str(key),
                    workspace=suite.workspace,
                ),
                workspace=suite.workspace,
            )
            suite.data["signing_keys"] = [key.fingerprint]
            suite.save()

        response = self.archive_file_get(path="")

        tree = self.assertResponseHTML(response)
        archive_suites = tree.xpath("//div[@id='archive-suites']")[0]
        self.assertTextContentEqual(
            archive_suites.p, "This archive contains the following suites:"
        )
        self.assertEqual(len(archive_suites.ul.li), len(self.suites))
        fqdn = settings.DEBUSINE_DEBIAN_ARCHIVE_FQDN
        workspace = self.scenario.workspace
        for li, suite, key in zip(archive_suites.ul.li, sorted_suites, keys):
            self.assertTextContentEqual(li.p, f"{suite.name}: ")
            suite_deb822 = Deb822("".join(li.pre.itertext()))
            self.assertCountEqual(
                suite_deb822.keys(), ["Types", "URIs", "Suites", "Signed-By"]
            )
            self.assertDictContainsAll(
                suite_deb822,
                {
                    "Types": "deb",
                    "URIs": (
                        f"http://{fqdn}/{workspace.scope.name}/{workspace.name}"
                    ),
                    "Suites": suite.name,
                },
            )
            self.assertEqual(
                PGPKeyring(
                    parse_multiline(suite_deb822["Signed-By"])
                ).fingerprints(),
                {key.fingerprint},
            )
        self.assert_has_caching_headers(response, mutable=True)

    def test_snapshot(self) -> None:
        now = timezone.now()
        one_month_ago = (now - dt.timedelta(days=30)).replace(microsecond=0)

        response = self.archive_file_get(snapshot=one_month_ago, path="")

        # We always list all exported suites in the workspace, regardless of
        # when they were added to the archive, so we don't bother testing
        # different snapshot times here.
        tree = self.assertResponseHTML(response)
        archive_suites = tree.xpath("//div[@id='archive-suites']")[0]
        self.assertTextContentEqual(
            archive_suites.p, "This archive contains the following suites:"
        )
        self.assertEqual(len(archive_suites.ul.li), len(self.suites))
        fqdn = settings.DEBUSINE_DEBIAN_ARCHIVE_FQDN
        workspace = self.scenario.workspace
        for li, suite in zip(
            archive_suites.ul.li,
            sorted(self.suites.values(), key=attrgetter("name")),
        ):
            self.assertTextContentEqual(li.p, f"{suite.name}: ")
            self.assertEqual(
                Deb822("".join(li.pre.itertext())),
                {
                    "Types": "deb",
                    "URIs": (
                        f"http://{fqdn}/{workspace.scope.name}/{workspace.name}"
                    ),
                    "Suites": suite.name,
                    "Snapshot": self.format_snapshot(one_month_ago),
                },
            )
        self.assert_has_caching_headers(response, mutable=False)

    def test_archive_empty(self) -> None:
        self.archive.child_items.all().delete()
        for suite in self.suites.values():
            suite.delete()

        for snapshot in (None, timezone.now()):
            with self.subTest(snapshot=self.format_snapshot(snapshot)):
                response = self.archive_file_get(snapshot=snapshot, path="")

                tree = self.assertResponseHTML(response)
                archive_suites = tree.xpath("//div[@id='archive-suites']")[0]
                self.assertTextContentEqual(
                    archive_suites, "This archive is empty."
                )
                self.assert_has_caching_headers(
                    response, mutable=snapshot is None
                )

    def test_get_place(self) -> None:
        response = self.archive_file_get(path="")
        assert response.context_data is not None

        place = response.context_data["place"]
        workspace = self.scenario.workspace

        self.assertEqual(place.title, f"Suites in <code>{workspace}</code>")

    @override_settings(
        ALLOWED_HOSTS=["*"],
        DEBUSINE_DEBIAN_ARCHIVE_FQDN=["deb.example.com", "deb.example.org"],
        DEBUSINE_DEBIAN_ARCHIVE_PRIMARY_FQDN="deb.example.com",
    )
    def test_override_archive_fqdn(self) -> None:
        response = self.archive_file_get(path="")

        tree = self.assertResponseHTML(response)
        archive_suites = tree.xpath("//div[@id='archive-suites']")[0]
        workspace = self.scenario.workspace
        for li in archive_suites.ul.li:
            self.assertEqual(
                Deb822("".join(li.pre.itertext()))["URIs"],
                f"http://deb.example.com/"
                f"{workspace.scope.name}/{workspace.name}",
            )


class SuiteRootViewTests(ArchiveViewTests):
    """Test retrieving the root of a suite."""

    example_path = "dists/bookworm/"
    expect_in_suite = True
    expect_mutable_no_archive = True
    expect_mutable = True

    def test_current(self) -> None:
        suite = self.suites["bookworm"]
        suite.data["components"] = ["main", "contrib", "non-free"]
        suite.data["architectures"] = ["amd64"]
        suite.data["release_fields"] = {"Description": "Something special"}
        suite.save()

        response = self.archive_file_get(path="dists/bookworm/")

        tree = self.assertResponseHTML(response)
        self.assertTextContentEqual(
            tree.xpath("//div[@id='suite-summary']")[0],
            "bookworm Something special",
        )

        self.assertTextContentEqual(
            tree.xpath("//div[@id='suite-architectures']")[0],
            "Architectures amd64",
        )

        suite_release_fields = tree.xpath("//div[@id='suite-release-fields']")[
            0
        ]
        self.assertEqual(
            [
                (element.tag, element.text.strip())
                for element in suite_release_fields.dl.iterchildren()
            ],
            [("dt", "Description"), ("dd", "Something special")],
        )
        suite_apt = tree.xpath("//div[@id='suite-apt']")[0]
        fqdn = settings.DEBUSINE_DEBIAN_ARCHIVE_FQDN
        workspace = self.scenario.workspace
        self.assertEqual(
            Deb822("".join(suite_apt.pre.itertext())),
            {
                "Types": "deb",
                "URIs": (
                    f"http://{fqdn}/{workspace.scope.name}/{workspace.name}"
                ),
                "Suites": "bookworm",
                "Components": "main contrib non-free",
            },
        )
        self.assert_has_caching_headers(response, mutable=True)

    def test_current_with_required_suites(self) -> None:
        personal = self.playground.create_collection(
            "r-personal",
            CollectionCategory.SUITE,
            workspace=self.scenario.workspace,
        )
        CollectionRelation.objects.create(
            source=personal,
            target=self.suites["trixie"],
            created_by_user=self.playground.get_default_user(),
            type=CollectionRelation.Types.SUITE_REQUIRES,
            position=1,
        )

        response = self.archive_file_get(path="dists/r-personal/")

        tree = self.assertResponseHTML(response)

        fqdn = settings.DEBUSINE_DEBIAN_ARCHIVE_FQDN
        workspace = self.scenario.workspace
        uri = f"http://{fqdn}/{workspace.scope.name}/{workspace.name}"

        required_apt_sources = "".join(
            tree.xpath("//div[@id='suite-required-apt']/pre/code/text()")
        )

        self.assertEqual(
            list(Deb822.iter_paragraphs(required_apt_sources)),
            [
                {
                    "Types": "deb",
                    "URIs": uri,
                    "Suites": "trixie",
                },
            ],
        )

    def test_current_signing_keys(self) -> None:
        suite = self.suites["bookworm"]
        keys = [
            PGPKey.new(PubKeyAlgorithm.EdDSA, EllipticCurveOID.Ed25519).pubkey
            for _ in range(2)
        ]
        for key in keys:
            self.playground.create_asset_usage(
                resource=self.playground.create_signing_key_asset(
                    purpose=KeyPurpose.OPENPGP_REPOSITORY,
                    fingerprint=key.fingerprint,
                    public_key=str(key),
                    workspace=suite.workspace,
                ),
                workspace=suite.workspace,
            )
        suite.data["signing_keys"] = [key.fingerprint for key in keys]
        suite.save()

        response = self.archive_file_get(path="dists/bookworm/")

        tree = self.assertResponseHTML(response)
        self.assertTextContentEqual(
            tree.xpath("//div[@id='suite-summary']")[0], "bookworm"
        )
        self.assertFalse(tree.xpath("//div[@id='suite-architectures']"))
        self.assertFalse(tree.xpath("//div[@id='suite-release-fields']"))
        fqdn = settings.DEBUSINE_DEBIAN_ARCHIVE_FQDN
        workspace = self.scenario.workspace
        suite_apt = tree.xpath("//div[@id='suite-apt']")[0]
        suite_deb822 = Deb822("".join(suite_apt.pre.itertext()))
        self.assertCountEqual(
            suite_deb822.keys(), ["Types", "URIs", "Suites", "Signed-By"]
        )
        self.assertDictContainsAll(
            suite_deb822,
            {
                "Types": "deb",
                "URIs": (
                    f"http://{fqdn}/{workspace.scope.name}/{workspace.name}"
                ),
                "Suites": "bookworm",
            },
        )
        self.assertEqual(
            PGPKeyring(
                parse_multiline(suite_deb822["Signed-By"])
            ).fingerprints(),
            {key.fingerprint for key in keys},
        )
        self.assert_has_caching_headers(response, mutable=True)

    def test_snapshot(self) -> None:
        now = timezone.now()
        one_month_ago = (now - dt.timedelta(days=30)).replace(microsecond=0)
        self.archive.child_items.filter(
            category=CollectionCategory.SUITE, name="bookworm"
        ).update(created_at=one_month_ago)

        # We always show an exported suite, regardless of when it was added
        # to the archive, so we don't bother testing different snapshot
        # times here.
        response = self.archive_file_get(
            snapshot=one_month_ago, path="dists/bookworm/"
        )

        tree = self.assertResponseHTML(response)
        self.assertTextContentEqual(
            tree.xpath("//div[@id='suite-summary']")[0], "bookworm"
        )
        self.assertFalse(tree.xpath("//div[@id='suite-architectures']"))
        self.assertFalse(tree.xpath("//div[@id='suite-release-fields']"))
        fqdn = settings.DEBUSINE_DEBIAN_ARCHIVE_FQDN
        suite_apt = tree.xpath("//div[@id='suite-apt']")[0]
        workspace = self.scenario.workspace
        self.assertEqual(
            Deb822("".join(suite_apt.pre.itertext())),
            {
                "Types": "deb",
                "URIs": (
                    f"http://{fqdn}/{workspace.scope.name}/{workspace.name}"
                ),
                "Suites": "bookworm",
                "Snapshot": self.format_snapshot(one_month_ago),
            },
        )
        self.assert_has_caching_headers(response, mutable=False)

    def test_suite_not_found(self) -> None:
        for snapshot in (None, timezone.now()):
            with self.subTest(snapshot=self.format_snapshot(snapshot)):
                response = self.archive_file_get(
                    snapshot=snapshot, path="dists/nonexistent/"
                )

                self.assertResponse404(response, "Suite not found")
                self.assert_has_caching_headers(
                    response, mutable=snapshot is None
                )

    def test_suite_not_exported(self) -> None:
        self.playground.create_collection(
            "not-exported",
            CollectionCategory.SUITE,
            workspace=self.scenario.workspace,
            data={"exported": False},
        )

        for snapshot in (None, timezone.now()):
            with self.subTest(snapshot=self.format_snapshot(snapshot)):
                response = self.archive_file_get(
                    snapshot=snapshot, path="dists/not-exported/"
                )

                self.assertResponse404(response, "Suite not found")
                self.assert_has_caching_headers(
                    response, mutable=snapshot is None
                )

    def test_get_place(self) -> None:
        suite = self.suites["bookworm"]

        response = self.archive_file_get(path="dists/bookworm/")
        assert response.context_data is not None

        place = response.context_data["place"]

        self.assertEqual(place.title, f"Suite <code>{suite}</code>")


class SigningKeysViewArchiveTests(ArchiveViewTests):
    """Test retrieving signing keys for an archive."""

    keys: ClassVar[dict[str, PGPKey]]

    example_path = "signing-keys.asc"
    expect_in_suite = False
    expect_mutable_no_archive = True
    expect_mutable = True
    expect_content_type = "application/pgp-keys; charset=utf-8"

    @override
    @classmethod
    def setUpTestData(cls) -> None:
        super().setUpTestData()
        cls.keys = {
            suite_name: PGPKey.new(
                PubKeyAlgorithm.EdDSA, EllipticCurveOID.Ed25519
            ).pubkey
            for suite_name in cls.suites
        }
        for suite_name, suite in cls.suites.items():
            cls.playground.create_asset_usage(
                resource=cls.playground.create_signing_key_asset(
                    purpose=KeyPurpose.OPENPGP_REPOSITORY,
                    fingerprint=cls.keys[suite_name].fingerprint,
                    public_key=str(cls.keys[suite_name]),
                    workspace=suite.workspace,
                ),
                workspace=suite.workspace,
            )
            suite.data["signing_keys"] = [cls.keys[suite_name].fingerprint]
            suite.save()

    def test_current(self) -> None:
        response = self.archive_file_get(path=self.example_path)

        self.assertEqual(response.status_code, status.HTTP_200_OK)
        self.assertEqual(
            response.headers["Content-Type"], self.expect_content_type
        )
        self.assertEqual(
            PGPKeyring(response.content).fingerprints(),
            {key.fingerprint for key in self.keys.values()},
        )
        self.assert_has_caching_headers(response, mutable=True)


class SigningKeysViewSuiteTests(ArchiveViewTests):
    """Test retrieving signing keys for a suite."""

    keys: ClassVar[list[PGPKey]]

    example_path = "dists/bookworm/signing-keys.asc"
    expect_in_suite = True
    expect_mutable_no_archive = True
    expect_mutable = True
    expect_content_type = "application/pgp-keys; charset=utf-8"

    @override
    @classmethod
    def setUpTestData(cls) -> None:
        super().setUpTestData()
        suite = cls.suites["bookworm"]
        cls.keys = [
            PGPKey.new(PubKeyAlgorithm.EdDSA, EllipticCurveOID.Ed25519).pubkey
            for _ in range(2)
        ]
        for key in cls.keys:
            cls.playground.create_asset_usage(
                resource=cls.playground.create_signing_key_asset(
                    purpose=KeyPurpose.OPENPGP_REPOSITORY,
                    fingerprint=key.fingerprint,
                    public_key=str(key),
                    workspace=suite.workspace,
                ),
                workspace=suite.workspace,
            )
        suite.data["signing_keys"] = [key.fingerprint for key in cls.keys]
        suite.save()

    def test_current(self) -> None:
        response = self.archive_file_get(path=self.example_path)

        self.assertEqual(response.status_code, status.HTTP_200_OK)
        self.assertEqual(
            response.headers["Content-Type"], self.expect_content_type
        )
        self.assertEqual(
            PGPKeyring(response.content).fingerprints(),
            {key.fingerprint for key in self.keys},
        )
        self.assert_has_caching_headers(response, mutable=True)


class DistsByHashFileViewTests(ArchiveFileViewTests):
    """Test retrieving ``by-hash`` files from ``dists/``."""

    contents: ClassVar[dict[str, bytes]]
    sha256: ClassVar[dict[str, str]]

    example_path_nonexistent = "dists/bookworm/main/source/by-hash/SHA256/00"
    expect_in_suite = True
    expect_mutable_no_archive = False
    expect_mutable = False
    expect_content_type = "application/x-xz"

    @override
    @classmethod
    def setUpTestData(cls) -> None:
        super().setUpTestData()
        cls.contents = {
            "old": b"Old Sources file\n",
            "new": b"New Sources file\n",
        }
        cls.sha256 = {
            name: hashlib.sha256(c).hexdigest()
            for name, c in cls.contents.items()
        }
        cls.suites["bookworm"].manager.add_artifact(
            cls.playground.create_repository_index(
                "Sources.xz", contents=cls.contents["old"]
            ),
            user=cls.scenario.user,
            variables={"path": "main/source/Sources.xz"},
        )
        cls.suites["bookworm"].manager.add_artifact(
            cls.playground.create_repository_index(
                "Sources.xz", contents=cls.contents["new"]
            ),
            user=cls.scenario.user,
            variables={"path": "main/source/Sources.xz"},
            replace=True,
        )
        cls.suites["bookworm"].manager.add_artifact(
            cls.playground.create_repository_index("Packages.xz"),
            user=cls.scenario.user,
            variables={"path": "main/binary-amd64/Packages.xz"},
        )
        cls.suites["trixie"].manager.add_artifact(
            cls.playground.create_repository_index("Sources.xz"),
            user=cls.scenario.user,
            variables={"path": "main/source/Sources.xz"},
        )
        cls.example_path = (
            f"dists/bookworm/main/source/by-hash/SHA256/{cls.sha256['new']}"
        )

    def test_current(self) -> None:
        response = self.archive_file_get(
            path=(
                f"dists/bookworm/main/source/by-hash/SHA256/"
                f"{self.sha256['old']}"
            )
        )

        self.assert_response_ok(
            response, contents=self.contents["old"], mutable=False
        )

        response = self.archive_file_get(
            path=(
                f"dists/bookworm/main/source/by-hash/SHA256/"
                f"{self.sha256['new']}"
            )
        )

        self.assert_response_ok(
            response, contents=self.contents["new"], mutable=False
        )

    def test_snapshot(self) -> None:
        one_month_ago = (timezone.now() - dt.timedelta(days=30)).replace(
            microsecond=0
        )
        self.archive.child_items.filter(
            category=CollectionCategory.SUITE
        ).update(created_at=one_month_ago)
        old_item = (
            self.suites["bookworm"]
            .child_items.filter(name="index:main/source/Sources.xz")
            .order_by("id")[0]
        )
        old_item.created_at = one_month_ago
        old_item.save()
        new_item = (
            self.suites["bookworm"]
            .child_items.filter(name="index:main/source/Sources.xz")
            .order_by("id")[1]
        )

        for snapshot, sha256, expected_contents in (
            (
                old_item.created_at - dt.timedelta(seconds=1),
                self.sha256["old"],
                None,
            ),
            (
                old_item.created_at - dt.timedelta(seconds=1),
                self.sha256["new"],
                None,
            ),
            (old_item.created_at, self.sha256["old"], self.contents["old"]),
            (old_item.created_at, self.sha256["new"], None),
            (
                new_item.created_at - dt.timedelta(seconds=1),
                self.sha256["old"],
                self.contents["old"],
            ),
            (
                new_item.created_at - dt.timedelta(seconds=1),
                self.sha256["new"],
                None,
            ),
            # Snapshot URLs don't include a microsecond component, so add a
            # second to avoid rounding problems.
            (
                new_item.created_at + dt.timedelta(seconds=1),
                self.sha256["old"],
                self.contents["old"],
            ),
            (
                new_item.created_at + dt.timedelta(seconds=1),
                self.sha256["new"],
                self.contents["new"],
            ),
        ):
            with self.subTest(
                snapshot=self.format_snapshot(snapshot), sha256=sha256
            ):
                response = self.archive_file_get(
                    snapshot=snapshot,
                    path=f"dists/bookworm/main/source/by-hash/SHA256/{sha256}",
                )

                if expected_contents is None:
                    self.assertResponse404(
                        response, "No FileInArtifact matches the given query."
                    )
                else:
                    self.assert_response_ok(
                        response, contents=expected_contents, mutable=False
                    )

    def test_suite_not_exported(self) -> None:
        response = self.archive_file_get(
            path=(
                f"dists/bookworm/main/source/by-hash/SHA256/"
                f"{self.sha256['new']}"
            )
        )

        self.assert_response_ok(
            response, contents=self.contents["new"], mutable=False
        )

        self.suites["bookworm"].data["exported"] = False
        self.suites["bookworm"].save()

        response = self.archive_file_get(
            path=(
                f"dists/bookworm/main/source/by-hash/SHA256/"
                f"{self.sha256['new']}"
            )
        )

        self.assertResponse404(
            response, "No FileInArtifact matches the given query."
        )
        self.assert_has_caching_headers(response, mutable=False)

    def test_ignores_content_negotiation(self) -> None:
        response = self.archive_file_get(
            path=(
                f"dists/bookworm/main/source/by-hash/SHA256/"
                f"{self.sha256['new']}"
            ),
            headers={"Accept": "application/octet-stream"},
        )

        self.assert_response_ok(
            response, contents=self.contents["new"], mutable=False
        )

    def test_by_hash_matches_multiple_artifacts(self) -> None:
        contents = b"Sources file"
        sha256 = hashlib.sha256(contents).hexdigest()
        for _ in range(2):
            self.suites["bookworm"].manager.add_artifact(
                self.playground.create_repository_index(
                    "Sources.xz", contents=contents
                ),
                replace=True,
                user=self.scenario.user,
                variables={"path": "main/source/Sources.xz"},
            )

        response = self.archive_file_get(
            path=f"dists/bookworm/main/source/by-hash/SHA256/{sha256}",
        )

        self.assert_response_ok(
            response, contents=b"Sources file", mutable=False
        )


class DistsFileViewTests(ArchiveFileViewTests):
    """Test retrieving non-``by-hash`` files from ``dists/``."""

    contents: ClassVar[dict[str, bytes]]

    example_path = "dists/bookworm/Release"
    example_path_nonexistent = "dists/bookworm/Nonexistent"
    expect_in_suite = True
    expect_mutable_no_archive = True
    expect_mutable = True
    # This is really just the Content-Type of the Release file.
    expect_content_type = "application/octet-stream"

    @override
    @classmethod
    def setUpTestData(cls) -> None:
        super().setUpTestData()
        cls.contents = {
            "old": b"Old Release file\n",
            "new": b"New Release file\n",
        }
        cls.suites["bookworm"].manager.add_artifact(
            cls.playground.create_repository_index(
                "Release", contents=b"Old Release file\n"
            ),
            user=cls.scenario.user,
            variables={"path": "Release"},
        )
        cls.suites["bookworm"].manager.add_artifact(
            cls.playground.create_repository_index(
                "Release", contents=b"New Release file\n"
            ),
            user=cls.scenario.user,
            variables={"path": "Release"},
            replace=True,
        )
        cls.suites["bookworm"].manager.add_artifact(
            cls.playground.create_repository_index("Sources.xz"),
            user=cls.scenario.user,
            variables={"path": "main/source/Sources.xz"},
        )
        cls.suites["trixie"].manager.add_artifact(
            cls.playground.create_repository_index("Release"),
            user=cls.scenario.user,
            variables={"path": "Release"},
        )

    def test_current(self) -> None:
        response = self.archive_file_get(path="dists/bookworm/Release")

        self.assert_response_ok(
            response, contents=self.contents["new"], mutable=True
        )

    def test_snapshot(self) -> None:
        one_month_ago = (timezone.now() - dt.timedelta(days=30)).replace(
            microsecond=0
        )
        self.archive.child_items.filter(
            category=CollectionCategory.SUITE
        ).update(created_at=one_month_ago)
        old_item = (
            self.suites["bookworm"]
            .child_items.filter(name="index:Release")
            .order_by("id")[0]
        )
        old_item.created_at = one_month_ago
        old_item.save()
        new_item = (
            self.suites["bookworm"]
            .child_items.filter(name="index:Release")
            .order_by("id")[1]
        )

        for snapshot, expected_contents in (
            (old_item.created_at - dt.timedelta(seconds=1), None),
            (old_item.created_at, self.contents["old"]),
            (
                new_item.created_at - dt.timedelta(seconds=1),
                self.contents["old"],
            ),
            # Snapshot URLs don't include a microsecond component, so add a
            # second to avoid rounding problems.
            (
                new_item.created_at + dt.timedelta(seconds=1),
                self.contents["new"],
            ),
        ):
            with self.subTest(snapshot=self.format_snapshot(snapshot)):
                response = self.archive_file_get(
                    snapshot=snapshot, path="dists/bookworm/Release"
                )

                if expected_contents is None:
                    self.assertResponse404(
                        response, "No FileInArtifact matches the given query."
                    )
                else:
                    self.assert_response_ok(
                        response, contents=expected_contents, mutable=False
                    )

    def test_suite_not_exported(self) -> None:
        response = self.archive_file_get(path="dists/bookworm/Release")

        self.assert_response_ok(
            response, contents=self.contents["new"], mutable=True
        )

        self.suites["bookworm"].data["exported"] = False
        self.suites["bookworm"].save()

        response = self.archive_file_get(path="dists/bookworm/Release")

        self.assertResponse404(
            response, "No FileInArtifact matches the given query."
        )
        self.assert_has_caching_headers(response, mutable=True)

    def test_ignores_content_negotiation(self) -> None:
        response = self.archive_file_get(
            path="dists/bookworm/Release",
            headers={"Accept": "application/octet-stream"},
        )

        self.assert_response_ok(
            response, contents=self.contents["new"], mutable=True
        )


class PoolFileViewTests(ArchiveFileViewTests):
    """Test retrieving files from ``pool/``."""

    example_path = "pool/contrib/h/hello/hello_1.1-1.dsc"
    example_path_nonexistent = "pool/main/n/nonexistent/nonexistent_1.0.dsc"
    expect_in_suite = True
    expect_mutable_no_archive = True
    expect_mutable = False
    expect_content_type = "text/plain; charset=utf-8"

    @override
    @classmethod
    def setUpTestData(cls) -> None:
        super().setUpTestData()
        for suite_name, component, name, version, paths in (
            (
                "bookworm",
                "main",
                "hello",
                "1.0",
                ["hello_1.0.dsc", "hello_1.0.tar.xz"],
            ),
            (
                "bookworm",
                "contrib",
                "hello",
                "1.1-1",
                [
                    "hello_1.1-1.dsc",
                    "hello_1.1-1.debian.tar.xz",
                    "hello_1.1.orig.tar.xz",
                ],
            ),
            (
                "bookworm",
                "contrib",
                "hello",
                "1.1-2",
                [
                    "hello_1.1-2.dsc",
                    "hello_1.1-2.debian.tar.xz",
                    "hello_1.1.orig.tar.xz",
                ],
            ),
            ("bookworm", "main", "base-files", "1.0", ["base-files_1.0.dsc"]),
            (
                "trixie",
                "main",
                "hello",
                "1.0",
                ["hello_1.0.dsc", "hello_1.0.tar.xz"],
            ),
        ):
            cls.suites[suite_name].manager.add_artifact(
                cls.playground.create_minimal_source_package_artifact(
                    name,
                    version,
                    paths={
                        path: f"Contents of {path}".encode() for path in paths
                    },
                    create_files=True,
                ),
                user=cls.scenario.user,
                variables={"component": component, "section": "devel"},
            )

        for (
            suite_name,
            component,
            srcpkg_name,
            srcpkg_version,
            name,
            version,
            architecture,
        ) in (
            ("bookworm", "main", "hello", "1.0", "libhello1", "1.0", "amd64"),
            ("bookworm", "main", "hello", "1.0", "libhello1", "1.0", "s390x"),
            ("bookworm", "main", "hello", "1.0", "libhello-doc", "1.0", "all"),
            (
                "bookworm",
                "contrib",
                "hello",
                "1.1",
                "libhello1",
                "1:1.1",
                "amd64",
            ),
            (
                "bookworm",
                "main",
                "base-files",
                "1.0",
                "base-files",
                "1.0",
                "amd64",
            ),
            ("trixie", "main", "hello", "1.0", "libhello1", "1.0", "amd64"),
        ):
            deb_path = f"{name}_{version.split(':')[-1]}_{architecture}.deb"
            cls.suites[suite_name].manager.add_artifact(
                cls.playground.create_minimal_binary_package_artifact(
                    srcpkg_name,
                    srcpkg_version,
                    name,
                    version,
                    architecture,
                    paths={deb_path: f"Contents of {deb_path}".encode()},
                    create_files=True,
                ),
                user=cls.scenario.user,
                variables={
                    "component": component,
                    "section": "devel",
                    "priority": "optional",
                },
            )

    def test_current_source(self) -> None:
        for path in (
            "pool/main/h/hello/hello_1.0.dsc",
            "pool/main/h/hello/hello_1.0.tar.xz",
            "pool/contrib/h/hello/hello_1.1-1.dsc",
            "pool/contrib/h/hello/hello_1.1-1.debian.tar.xz",
            "pool/contrib/h/hello/hello_1.1-2.dsc",
            "pool/contrib/h/hello/hello_1.1-2.debian.tar.xz",
            "pool/contrib/h/hello/hello_1.1.orig.tar.xz",
        ):
            response = self.archive_file_get(path=path)

            self.assert_response_ok(
                response,
                contents=f"Contents of {PurePath(path).name}".encode(),
                mutable=False,
            )

    def test_current_binary(self) -> None:
        for path in (
            "pool/main/h/hello/libhello1_1.0_amd64.deb",
            "pool/main/h/hello/libhello1_1.0_s390x.deb",
            "pool/main/h/hello/libhello-doc_1.0_all.deb",
            "pool/contrib/h/hello/libhello1_1.1_amd64.deb",
        ):
            response = self.archive_file_get(path=path)

            self.assert_response_ok(
                response,
                contents=f"Contents of {PurePath(path).name}".encode(),
                mutable=False,
            )

    def test_snapshot_source(self) -> None:
        one_month_ago = (timezone.now() - dt.timedelta(days=30)).replace(
            microsecond=0
        )
        one_day_ago = (timezone.now() - dt.timedelta(days=1)).replace(
            microsecond=0
        )
        self.archive.child_items.filter(
            category=CollectionCategory.SUITE
        ).update(created_at=one_month_ago)
        items = []
        for version in ("1.1-1", "1.1-2", "1.1-3"):
            paths = {
                path: f"Contents of {path}".encode()
                for path in (
                    f"hello-snapshot_{version}.dsc",
                    f"hello-snapshot_{version.split('-')[0]}.orig.tar.xz",
                )
            }
            items.append(
                self.suites["bookworm"].manager.add_artifact(
                    self.playground.create_minimal_source_package_artifact(
                        "hello-snapshot",
                        version,
                        paths=paths,
                        create_files=True,
                    ),
                    user=self.scenario.user,
                    variables={"component": "main", "section": "devel"},
                )
            )
        items[0].created_at = one_month_ago
        items[0].removed_at = one_day_ago
        items[0].save()
        items[1].created_at = one_day_ago
        items[1].save()
        start = items[0].created_at - dt.timedelta(seconds=1)
        # Snapshot URLs don't include a microsecond component, so add a
        # second to avoid rounding problems.
        end = items[2].created_at + dt.timedelta(seconds=1)

        for snapshot, name, expected in (
            (start, "hello-snapshot_1.1-1.dsc", False),
            (start, "hello-snapshot_1.1-2.dsc", False),
            (start, "hello-snapshot_1.1-3.dsc", False),
            (start, "hello-snapshot_1.1.orig.tar.xz", False),
            (items[0].created_at, "hello-snapshot_1.1-1.dsc", True),
            (items[0].created_at, "hello-snapshot_1.1-2.dsc", False),
            (items[0].created_at, "hello-snapshot_1.1-3.dsc", False),
            (items[0].created_at, "hello-snapshot_1.1.orig.tar.xz", True),
            (items[1].created_at, "hello-snapshot_1.1-1.dsc", False),
            (items[1].created_at, "hello-snapshot_1.1-2.dsc", True),
            (items[1].created_at, "hello-snapshot_1.1-3.dsc", False),
            (items[1].created_at, "hello-snapshot_1.1.orig.tar.xz", True),
            (end, "hello-snapshot_1.1-1.dsc", False),
            (end, "hello-snapshot_1.1-2.dsc", True),
            (end, "hello-snapshot_1.1-3.dsc", True),
            (end, "hello-snapshot_1.1.orig.tar.xz", True),
        ):
            with self.subTest(
                snapshot=self.format_snapshot(snapshot), name=name
            ):
                response = self.archive_file_get(
                    snapshot=snapshot, path=f"pool/main/h/hello-snapshot/{name}"
                )

                if expected:
                    self.assert_response_ok(
                        response,
                        contents=f"Contents of {name}".encode(),
                        mutable=False,
                    )
                else:
                    self.assertResponse404(
                        response, "No FileInArtifact matches the given query."
                    )

    def test_suite_not_exported(self) -> None:
        response = self.archive_file_get(path=self.example_path)

        self.assert_response_ok(
            response,
            contents=f"Contents of {self.example_path.split('/')[-1]}".encode(),
            mutable=False,
        )

        self.suites["bookworm"].data["exported"] = False
        self.suites["bookworm"].save()

        response = self.archive_file_get(path=self.example_path)

        self.assertResponse404(
            response, "No FileInArtifact matches the given query."
        )
        self.assert_has_caching_headers(response, mutable=False)

    def test_wrong_component(self) -> None:
        for name in ("hello_1.0.dsc", "libhello1_1.0_amd64.deb"):
            response = self.archive_file_get(path=f"pool/main/h/hello/{name}")

            self.assertEqual(response.status_code, status.HTTP_200_OK)

            response = self.archive_file_get(
                path=f"pool/contrib/h/hello/{name}"
            )

            self.assertResponse404(
                response, "No FileInArtifact matches the given query."
            )

    def test_wrong_source_prefix(self) -> None:
        for name in ("hello_1.0.dsc", "libhello1_1.0_amd64.deb"):
            response = self.archive_file_get(path=f"pool/main/h/hello/{name}")

            self.assertEqual(response.status_code, status.HTTP_200_OK)

            response = self.archive_file_get(path=f"pool/main/a/hello/{name}")

            self.assertResponse404(
                response, "No FileInArtifact matches the given query."
            )

    def test_ignores_content_negotiation(self) -> None:
        response = self.archive_file_get(
            path="pool/main/h/hello/hello_1.0.dsc",
            headers={"Accept": "application/octet-stream"},
        )

        self.assert_response_ok(
            response, contents=b"Contents of hello_1.0.dsc", mutable=False
        )


class TopLevelFileViewTests(ArchiveFileViewTests):
    """Test retrieving top-level files, not in any suite."""

    contents: ClassVar[dict[str, bytes]]

    example_path = "doc/social-contract.txt"
    example_path_nonexistent = "NONEXISTENT"
    expect_in_suite = False
    expect_mutable_no_archive = True
    expect_mutable = True
    expect_content_type = "text/plain; charset=utf-8"

    @override
    @classmethod
    def setUpTestData(cls) -> None:
        super().setUpTestData()
        cls.contents = {
            "old": b'"Social Contract" with the Free Software Community\n',
            "new": b'"Social Contract" with the Free Software Community (v2)\n',
        }
        old_item = cls.archive.manager.add_artifact(
            cls.playground.create_repository_index(
                "social-contract.txt", contents=cls.contents["old"]
            ),
            user=cls.scenario.user,
            variables={"path": "doc/social-contract.txt"},
        )
        old_item.created_at = (timezone.now() - dt.timedelta(days=30)).replace(
            microsecond=0
        )
        old_item.save()
        cls.archive.manager.add_artifact(
            cls.playground.create_repository_index(
                "social-contract.txt", contents=cls.contents["new"]
            ),
            user=cls.scenario.user,
            variables={"path": "doc/social-contract.txt"},
            replace=True,
        )

    def test_current(self) -> None:
        response = self.archive_file_get(path="doc/social-contract.txt")

        self.assert_response_ok(
            response, contents=self.contents["new"], mutable=True
        )

    def test_snapshot(self) -> None:
        old_item = self.archive.child_items.filter(
            name="index:doc/social-contract.txt"
        ).order_by("id")[0]
        new_item = self.archive.child_items.filter(
            name="index:doc/social-contract.txt"
        ).order_by("id")[1]

        for snapshot, expected_contents in (
            (old_item.created_at - dt.timedelta(seconds=1), None),
            (old_item.created_at, self.contents["old"]),
            (
                new_item.created_at - dt.timedelta(seconds=1),
                self.contents["old"],
            ),
            # Snapshot URLs don't include a microsecond component, so add a
            # second to avoid rounding problems.
            (
                new_item.created_at + dt.timedelta(seconds=1),
                self.contents["new"],
            ),
        ):
            with self.subTest(snapshot=self.format_snapshot(snapshot)):
                response = self.archive_file_get(
                    snapshot=snapshot, path="doc/social-contract.txt"
                )

                if expected_contents is None:
                    self.assertResponse404(
                        response, "No FileInArtifact matches the given query."
                    )
                else:
                    self.assert_response_ok(
                        response, contents=expected_contents, mutable=False
                    )

    def test_ignores_content_negotiation(self) -> None:
        response = self.archive_file_get(
            path=self.example_path,
            headers={"Accept": "application/octet-stream"},
        )

        self.assert_response_ok(
            response, contents=self.contents["new"], mutable=True
        )


class RenderableSuiteTests(TestCase):
    """Tests for RenderableSuite."""

    @override
    def setUp(self) -> None:
        super().setUp()
        self.trixie = self.playground.create_collection(
            "trixie", CollectionCategory.SUITE
        )

    def test_as_deb822_source(self) -> None:
        renderable = RenderableSuite(
            request_scheme="http", collection=self.trixie
        )
        self.assertEqual(
            renderable.as_deb822_source(),
            "Types: deb\n"
            "URIs: http://127.0.0.1:8001/debusine/System\n"
            "Suites: trixie\n",
        )

    def test_required_suites_as_deb822_sources(self) -> None:
        personal = self.playground.create_collection(
            "personal", CollectionCategory.SUITE
        )
        trixie_backports = self.playground.create_collection(
            "trixie-backports", CollectionCategory.SUITE
        )
        trixie_security = self.playground.create_collection(
            "trixie-security", CollectionCategory.SUITE
        )
        CollectionRelation.objects.create(
            source=personal,
            target=trixie_backports,
            type=CollectionRelation.Types.SUITE_REQUIRES,
            created_by_user=self.playground.get_default_user(),
            position=1,
        )
        CollectionRelation.objects.create(
            source=personal,
            target=trixie_security,
            type=CollectionRelation.Types.SUITE_REQUIRES,
            created_by_user=self.playground.get_default_user(),
            position=2,
        )
        CollectionRelation.objects.create(
            source=trixie_backports,
            target=self.trixie,
            type=CollectionRelation.Types.SUITE_REQUIRES,
            created_by_user=self.playground.get_default_user(),
            position=1,
        )
        CollectionRelation.objects.create(
            source=trixie_backports,
            target=trixie_security,
            type=CollectionRelation.Types.SUITE_REQUIRES,
            created_by_user=self.playground.get_default_user(),
            position=1,
        )

        trixie_backports_renderable = RenderableSuite(
            request_scheme="http", collection=trixie_backports
        )
        trixie_security_renderable = RenderableSuite(
            request_scheme="http", collection=trixie_security
        )
        trixie_renderable = RenderableSuite(
            request_scheme="http", collection=self.trixie
        )

        renderable = RenderableSuite(request_scheme="http", collection=personal)

        self.assertEqual(
            renderable.required_suites_as_deb822_sources(),
            f"{trixie_backports_renderable.as_deb822_source()}\n"
            f"{trixie_security_renderable.as_deb822_source()}\n"
            f"{trixie_renderable.as_deb822_source()}",
        )


# Avoid running tests from common base classes.
del ArchiveViewTests
del ArchiveFileViewTests



# Copyright © The Debusine Developers
# See the AUTHORS file at the top-level directory of this distribution
#
# This file is part of Debusine. It is subject to the license terms
# in the LICENSE file found in the top-level directory of this
# distribution. No part of Debusine, including this file, may be copied,
# modified, propagated, or distributed except according to the terms
# contained in the LICENSE file.

"""Workflow that creates a child workspace from the one it's run on."""

from typing import override

from debusine.server.tasks.models import CreateChildWorkspaceData
from debusine.server.workflows.base import Workflow
from debusine.server.workflows.models import (
    ChildWorkspaceData,
    WorkRequestWorkflowData,
)
from debusine.tasks.models import BaseDynamicTaskData


class CreateChildWorkspaceWorkflow(
    Workflow[ChildWorkspaceData, BaseDynamicTaskData]
):
    """
    Workflow that creates a new child workspace.

    This workflow is a simple interface for users to invoke the
    create_child_workspace server task.
    """

    TASK_NAME = "create_child_workspace"

    @override
    def populate(self) -> None:
        """Create the server task that creates the new workspace."""
        task_data = CreateChildWorkspaceData(
            prefix=self.data.prefix or self.workspace.name,
            suffix=self.data.suffix,
            public=self.data.public,
            owner_group=self.data.owner_group,
            workflow_template_names=self.data.workflow_template_names,
            expiration_delay=self.data.expiration_delay,
        )
        self.work_request_ensure_child_server(
            task_name="create_child_workspace",
            task_data=task_data,
            workflow_data=WorkRequestWorkflowData(
                display_name="Create child workspace",
                step="create-child-workspace",
            ),
        )

    @override
    def compute_dynamic_data(self) -> BaseDynamicTaskData:
        """
        Compute dynamic data for this workflow.

        :subject: suffix
        :parameter_summary: suffix
        """
        return BaseDynamicTaskData(
            subject=self.data.suffix, parameter_summary=self.data.suffix
        )



# Copyright © The Debusine Developers
# See the AUTHORS file at the top-level directory of this distribution
#
# This file is part of Debusine. It is subject to the license terms
# in the LICENSE file found in the top-level directory of this
# distribution. No part of Debusine, including this file, may be copied,
# modified, propagated, or distributed except according to the terms
# contained in the LICENSE file.

"""Tabular presentation of QuerySets."""

from collections.abc import Sequence
from functools import cached_property
from typing import Any, ClassVar, TYPE_CHECKING, cast, override

from django.core.exceptions import ImproperlyConfigured
from django.db.models import Model, QuerySet
from django.forms import Form
from django.http import HttpRequest

from debusine.utils import extract_generic_type_arguments
from debusine.web.forms import ModelFormBase
from debusine.web.views.table.columns import BoundColumn, Column, Columns
from debusine.web.views.table.declarative import Attribute
from debusine.web.views.table.filters import Filter, Filters
from debusine.web.views.table.ordering import Ordering, TableOrderings

if TYPE_CHECKING:
    from django.core.paginator import _SupportsPagination

    from debusine.web.views.table.paginator import Paginator

    _SupportsPagination  # fake usage for vulture


class Table[M: Model]:
    """Definition of a tabular presentation of a QuerySet."""

    #: Model class
    model: type[M]

    #: Columns declaratively defined in subclass definitions,
    #: collected by __init_subclass__
    column_definitions: ClassVar[dict[str, Column]]

    #: Orderings declaratively defined in subclass definitions,
    #: collected by __init_subclass__
    ordering_definitions: ClassVar[dict[str, Ordering]]

    #: Filters declaratively defined in subclass definitions,
    #: collected by __init_subclass__
    filter_definitions: ClassVar[dict[str, Filter]]

    #: Form class for a form to use for filtering results
    filter_form_class: type[Form] | type[ModelFormBase[M]] | None = None

    #: Default sorting for the table
    default_order: str

    #: Template name to use to render the table as a widget
    template_name: str | None = None

    #: The column selected for sorting
    current_column: BoundColumn[M]

    #: Related fields to fetch in the same query, for performance
    related_fields: Sequence[str] | None = None

    @classmethod
    def _collect_declarative[Attr: Attribute](
        cls,
        attr_class: type[Attr],
        collected_name: str,
        *,
        prefix: str | None = None,
    ) -> None:
        """
        Collect declaratively defined members.

        :param attr_class: members need to be an instance of this class
        :param collected_name: class member with a dictionary of collected
                               members, indexed by name
        :param prefix: if defined, enforce that member names begin with this
                       prefix. This only applies to names used for class
                       members: names defined inside member instances can be
                       anything
        """
        # We cannot use inspect.getmembers as it sorts results by name, and we
        # need to preserve the order of definitions

        definitions: dict[str, Attr] = {}

        # Add definitions from base classes
        for base in cls.__bases__:
            if not issubclass(base, Table):
                continue
            definitions.update(getattr(base, collected_name, {}))

        # Add definitions from class dict
        for name, member in cls.__dict__.items():
            if not isinstance(member, attr_class):
                continue

            # Enforce member prefix
            if prefix is not None:
                if not name.startswith(prefix):
                    raise ImproperlyConfigured(
                        f"{name} = {member.__class__.__name__}(…):"
                        f" assigned name does not start with {prefix!r}"
                    )
                name = name.removeprefix(prefix)

            # Set the default name
            if not hasattr(member, "name"):
                member.name = name

            # Store the definition using the member's name
            definitions[member.name] = member

        # Store the collected members
        setattr(cls, collected_name, definitions)

    @override
    def __init_subclass__(cls, **kwargs: Any) -> None:
        """Collect column and filter definitions."""
        super().__init_subclass__(**kwargs)
        [cls.model] = extract_generic_type_arguments(cls, Table)

        cls._collect_declarative(Column, "column_definitions")
        # Mypy disallow using type[T] where T is an abstract class.
        # However, here we are not instantiating, only using it for isinstance
        # checks, so we silence the warning.
        # See https://github.com/python/mypy/issues/4717
        cls._collect_declarative(
            Filter,  # type: ignore[type-abstract]
            "filter_definitions",
            prefix="filter_",
        )
        cls._collect_declarative(
            Ordering,
            "ordering_definitions",
            prefix="ordering_",
        )

    def __init__(
        self,
        request: HttpRequest,
        object_list: "_SupportsPagination[M]",
        *,
        prefix: str = "",
        default_order: str | None = None,
        preview: bool = False,
        fragment: str | None = None,
    ) -> None:
        """
        Build the table definition.

        :param request: the current request object
        :param object_list: the queryset to present
        :param prefix: the query string prefix for paginator arguments (set
                       this if you paginate multiple object lists in the same
                       page)
        :param default_order: name of the default column to use for sorting.
                               Prefix with '-' to make it descending. Use
                               ``colname.sortname`` to specify the sort option
                               by name.
        :param preview: if True, activate preview mode: filtering and ordering
                        are disabled, only the first page is shown, and the
                        page navigation is replaced by a small summary
        :param fragment: if set, append ``#fragment`` to generated URLs
        """
        #: Current request
        self.request = request

        if not isinstance(object_list, QuerySet):
            raise ImproperlyConfigured(
                "Table cannot work with an object_list"
                f" of class {object_list.__class__.__name__},"
                " only with QuerySets"
            )
        #: Queryset to filter
        self.queryset: QuerySet[M, M] = cast(QuerySet[M, M], object_list)

        #: Query string argument prefix
        self.prefix = f"{prefix}-" if prefix else ""

        #: ID for the table in the DOM
        #: It also serves as a base for DOM IDs for all table elements
        self.dom_id = f"table-{self.prefix}"

        #: URL with the full table if we are a preview
        # TODO: allow to subclass Table to specify a custom widget for the
        # footer in preview mode. Example use case: link to the view with the
        # full table
        self.is_preview = preview

        #: Optional fragment to append to generated URLs
        self.fragment = fragment

        # Override definition default with user-provided column name
        if default_order is not None:
            self.default_order = default_order

        #: Column definitions
        self.columns: Columns[M] = Columns(self)

        #: Filters
        self.filters: Filters[M] = Filters(self)

        #: Orderings
        self.orderings: TableOrderings[M] = TableOrderings(self)

        # Allow subclasses to programmatically add to the table definition
        self.init()

        self._post_init()

    def init(self) -> None:
        """
        Set up Table attributes.

        This function does nothing by default, and can be overridden by
        subclasses to customize the paginator for their specific use.
        """
        # Instantiate declaratively defined columns
        for name, column in self.column_definitions.items():
            self.add_column(name, column)

        # Instantiate declaratively defined orderings
        for name, ordering in self.ordering_definitions.items():
            self.add_ordering(name, ordering)

        # Instantiate declaratively defined filters
        for filter_ in self.filter_definitions.values():
            self.add_filter(filter_)

        if not self.is_preview:
            self.init_filters()

    def init_filters(self) -> None:
        """
        Let subclasses initialize table filters.

        This is not called if the table is in preview mode.
        """

    def add_column(self, name: str, column: Column) -> None:
        """Add a column to the table."""
        bc = self.columns.add(name, column)
        if column.ordering:
            column.ordering.name = name
            added = self.orderings.add(column.name, column.ordering)
            bc.orderings += added

    def add_ordering(self, name: str, ordering: Ordering) -> None:
        """Add an ordering to the table."""
        self.orderings.add(name, ordering)

    def add_filter(self, filter_: Filter) -> None:
        """Add a filter to the table."""
        if self.is_preview:
            return
        self.filters.add(filter_)

    def _post_init(self) -> None:
        """Finish setting up after the table definition is complete."""
        if (
            self.orderings.options
            and getattr(self, "default_order", None) is None
        ):
            raise ImproperlyConfigured(
                "table is sortable,"
                " but default_order not set in class or constructor arguments"
            )
        self.orderings.finalize()

    def get_paginator(
        self,
        per_page: int,
        *,
        orphans: int = 0,
        allow_empty_first_page: bool = True,
    ) -> "Paginator[M]":
        """Return the Paginator for this table."""
        from debusine.web.views.table.paginator import Paginator

        return Paginator(
            self,
            per_page=per_page,
            orphans=orphans,
            allow_empty_first_page=allow_empty_first_page,
        )

    @cached_property
    def rows(self) -> QuerySet[M, M]:
        """Return the queryset sorted as the user requested."""
        queryset = self.filters.filter_queryset(self.queryset)
        if self.related_fields:
            queryset = queryset.select_related(*self.related_fields)
        if self.orderings.options:
            queryset = self.orderings.order_queryset(queryset)
        return queryset

    def make_link(self, **kwargs: Any) -> str:
        """
        Make a link to the page, with query string elements changed.

        :param kwargs: key/value pairs to change in the URL. Keys are
          automatically prepended with table_prefix
        """
        query = self.request.GET.copy()
        for key, value in kwargs.items():
            query[self.prefix + key] = value
        res = query.urlencode()
        if self.fragment is not None:
            res += f"#{self.fragment}"
        return res



# Copyright © The Debusine Developers
# See the AUTHORS file at the top-level directory of this distribution
#
# This file is part of Debusine. It is subject to the license terms
# in the LICENSE file found in the top-level directory of this
# distribution. No part of Debusine, including this file, may be copied,
# modified, propagated, or distributed except according to the terms
# contained in the LICENSE file.

"""Base for views using the rightbar UI layout."""

from typing import Any, override

from django.views.generic.base import ContextMixin

from debusine.web.views.sidebar import SidebarItem


class RightbarUIMixin(ContextMixin):
    """Base class for views using a sidebar layout."""

    base_template = "web/_base_rightbar.html"

    def get_sidebar_items(self) -> list[SidebarItem]:
        """Return a list of sidebar items."""
        return []

    @override
    def get_context_data(self, **kwargs: Any) -> dict[str, Any]:
        """Add UI elements to the template context."""
        context = super().get_context_data(**kwargs)
        context["sidebar_items"] = self.get_sidebar_items()
        return context



# Copyright © The Debusine Developers
# See the AUTHORS file at the top-level directory of this distribution
#
# This file is part of Debusine. It is subject to the license terms
# in the LICENSE file found in the top-level directory of this
# distribution. No part of Debusine, including this file, may be copied,
# modified, propagated, or distributed except according to the terms
# contained in the LICENSE file.

"""Data models for assets."""

import enum
from collections.abc import Generator, Iterable
from typing import Any, Self, TYPE_CHECKING, TypeAlias, cast, override

from django.core.exceptions import ImproperlyConfigured, ValidationError
from django.db import models
from django.db.models import Exists, OuterRef, Q
from django.db.models.constraints import CheckConstraint, UniqueConstraint

from debusine.assets import (
    AssetCategory,
    BaseAssetDataModel,
    BasicAPTAuthenticationData,
    SigningKeyData,
    asset_data_model,
)
from debusine.db.constraints import JsonDataUniqueConstraint
from debusine.db.models import permissions
from debusine.db.models.permissions import (
    Allow,
    Role,
    enforce,
    permission_check,
    permission_filter,
)
from debusine.db.models.scopes import Scope
from debusine.db.models.workspaces import Workspace
from debusine.db.permissioncontext import PermissionContext
from debusine.tasks.models import WorkerType
from debusine.utils.typing_utils import copy_signature_from

if TYPE_CHECKING:
    from django_stubs_ext.db.models import TypedModelMeta
else:
    TypedModelMeta = object


class AssetRoles(permissions.Roles, permissions.RoleBase, enum.ReprEnum):
    """Available roles for an Asset."""

    OWNER = "owner"


AssetRoles.setup()


class AssetUsageRoleBase(permissions.RoleBase):
    """AssetUsage role implementation."""

    implied_by_scope_roles: frozenset[Scope.Roles]
    implied_by_workspace_roles: frozenset[Workspace.Roles]
    implied_by_asset_usage_roles: frozenset["AssetUsageRoles"]

    @override
    def _setup(self) -> None:
        """Set up implications for a newly constructed role."""
        implied_by_scope_roles: set[Scope.Roles] = set()
        implied_by_workspace_roles: set[Workspace.Roles] = set()
        implied_by_asset_usage_roles: set[AssetUsageRoles] = {
            cast(AssetUsageRoles, self)
        }
        for i in self.implied_by:
            match i:
                case Workspace.Roles():
                    implied_by_scope_roles |= i.implied_by_scope_roles
                    implied_by_workspace_roles |= i.implied_by_workspace_roles
                case Role():
                    # Resolve a role passed during class definition into its
                    # enum instance
                    role = self.__class__(i.value)
                    implied_by_scope_roles |= role.implied_by_scope_roles
                    implied_by_workspace_roles |= (
                        role.implied_by_workspace_roles
                    )
                    implied_by_asset_usage_roles |= (
                        role.implied_by_asset_usage_roles
                    )
                case _:
                    raise ImproperlyConfigured(
                        f"AssetUsage roles do not support implications by {i!r}"
                    )
        self.implied_by_scope_roles = frozenset(implied_by_scope_roles)
        self.implied_by_workspace_roles = frozenset(implied_by_workspace_roles)
        self.implied_by_asset_usage_roles = frozenset(
            implied_by_asset_usage_roles
        )

    def q(self, pc: PermissionContext) -> Q:
        """Return a Q expression to select asset usages with this role."""
        q = Q(
            roles__group__in=pc.groups(),
            roles__role__in=self.implied_by_asset_usage_roles,
        )
        if self.implied_by_workspace_roles:
            q |= Q(
                workspace__in=Workspace.objects.filter(
                    Q(
                        roles__group__in=pc.groups(),
                        roles__role__in=self.implied_by_workspace_roles,
                    )
                    | Q(
                        scope__in=Scope.objects.filter(
                            roles__group__in=pc.groups(),
                            roles__role__in=self.implied_by_scope_roles,
                        )
                    )
                )
            )
        return q

    def implies(self, role: "AssetUsageRoles") -> bool:
        """Check if this role implies the given one."""
        return (
            self.implied_by_scope_roles <= role.implied_by_scope_roles
            and self.implied_by_workspace_roles
            <= role.implied_by_workspace_roles
            and self.implied_by_asset_usage_roles
            <= role.implied_by_asset_usage_roles
        )


class AssetUsageRoles(permissions.Roles, AssetUsageRoleBase, enum.ReprEnum):
    """Available roles for an AssetUsage."""

    SIGNER = Role(
        "signer",
        description="Sign data using this key",
    )

    REPOSITORY_SIGNER = Role(
        "repository_signer",
        label="Repository signer",
        implied_by=[SIGNER, Workspace.Roles.OWNER],
        description="Sign repositories using this key",
    )

    APT_AUTHENTICATOR = Role(
        "apt_authenticator",
        label="Can use APT authentication",
        implied_by=[Workspace.Roles.OWNER],
        description="Authenticate to an external APT repository "
        "using these credentials",
    )


AssetUsageRoles.setup()


class AssetQuerySet[A](models.QuerySet["Asset", A]):
    """Custom QuerySet for Asset."""

    def in_current_scope(self) -> "AssetQuerySet[A]":
        """Filter to assets in the current scope."""
        from debusine.db.context import context

        return self.filter(workspace__scope=context.require_scope())

    @permission_filter(work_request=Allow.PASS, anonymous=Allow.PASS)
    def can_display(self, pc: PermissionContext) -> "AssetQuerySet[A]":
        """Keep only Assets that can be displayed."""
        # Delegate to workspace can_display check
        return self.filter(
            workspace__in=Workspace.objects.can_display(pc)
        ).exclude(category=AssetCategory.CLOUD_PROVIDER_ACCOUNT)

    @permission_filter()
    def can_manage_permissions(
        self, pc: PermissionContext
    ) -> "AssetQuerySet[A]":
        """Filter to Assets that can be managed by user."""
        assert pc.user.is_authenticated
        return self.filter(
            roles__group__users=pc.user, roles__role=AssetRoles.OWNER
        )


class AssetManager(models.Manager["Asset"]):
    """Manager for the Asset model."""

    def get_roles_model(self) -> type["AssetRole"]:
        """Get the model used for role assignment."""
        return AssetRole

    @override
    def get_queryset(self) -> AssetQuerySet[Any]:
        """Use the custom QuerySet."""
        return AssetQuerySet(self.model, using=self._db)

    def get_by_slug(
        self, category: str, slug: str, workspace: Workspace | None = None
    ) -> "Asset":
        """Return an asset with a matching slug."""
        match category:
            case AssetCategory.SIGNING_KEY:
                purpose, fingerprint = slug.split(":", 1)
                return self.get(
                    category=category,
                    data__purpose=purpose,
                    data__fingerprint=fingerprint,
                )
            case AssetCategory.APT_AUTHENTICATION:
                assets = self.filter(category=category)
                if slug.startswith(":"):
                    return assets.get(
                        workspace__isnull=True, data__name=slug[1:]
                    )
                if ":" in slug:
                    assert workspace is not None
                    workspace_name, name = slug.split(":", 1)
                    return assets.get(
                        workspace__scope=workspace.scope,
                        workspace__name=workspace_name,
                        data__name=name,
                    )
                else:
                    return assets.get(workspace=workspace, data__name=slug)
            case _:
                raise ValueError(f"No slug defined for category '{category}'")


class AssetUsageQuerySet[A](models.QuerySet["AssetUsage", A]):
    """Custom QuerySet for AssetUsage."""

    def with_role(self, pc: PermissionContext, role: AssetUsageRoles) -> Self:
        """Keep only resources where the user has the given role."""
        if not pc.user.is_authenticated:
            return self.none()
        return self.filter(
            Exists(self.model.objects.filter(role.q(pc), pk=OuterRef("pk")))
        )

    @permission_filter(work_request=Allow.PASS)
    def can_sign_with(self, pc: PermissionContext) -> "AssetUsageQuerySet[A]":
        """Keep only AssetUsages that the user can sign with."""
        return self.with_role(pc, AssetUsageRoles.SIGNER)

    @permission_filter(work_request=Allow.PASS)
    def can_sign_repository_with(
        self, pc: PermissionContext
    ) -> "AssetUsageQuerySet[A]":
        """Keep only AssetUsages that the user can sign a repository with."""
        return self.with_role(pc, AssetUsageRoles.REPOSITORY_SIGNER)

    @permission_filter(work_request=Allow.PASS)
    def can_use_apt_authentication_with(
        self, pc: PermissionContext
    ) -> "AssetUsageQuerySet[A]":
        """Keep only AssetUsages that the user can use for APT auth."""
        return self.with_role(pc, AssetUsageRoles.APT_AUTHENTICATOR)


class AssetUsageManager(models.Manager["AssetUsage"]):
    """Manager for the AssetUsage model."""

    def get_roles_model(self) -> type["AssetUsageRole"]:
        """Get the model used for role assignment."""
        return AssetUsageRole


class Asset(models.Model):
    """Asset model."""

    category = models.CharField(max_length=255, choices=AssetCategory.choices)
    workspace = models.ForeignKey(
        Workspace, on_delete=models.PROTECT, blank=True, null=True
    )
    data = models.JSONField(default=dict)
    created_at = models.DateTimeField(auto_now_add=True)
    created_by = models.ForeignKey(
        "User", blank=True, null=True, on_delete=models.PROTECT
    )
    created_by_work_request = models.ForeignKey(
        "WorkRequest", blank=True, null=True, on_delete=models.SET_NULL
    )

    Roles: TypeAlias = AssetRoles
    objects = AssetManager.from_queryset(AssetQuerySet)()

    class Meta(TypedModelMeta):
        base_manager_name = "objects"
        constraints = [
            JsonDataUniqueConstraint(
                fields=["data->>'name'"],
                condition=models.Q(
                    category=AssetCategory.CLOUD_PROVIDER_ACCOUNT
                ),
                nulls_distinct=False,
                name="%(app_label)s_%(class)s_unique_cloud_provider_acct_name",
            ),
            JsonDataUniqueConstraint(
                fields=["data->>'fingerprint'"],
                condition=models.Q(category=AssetCategory.SIGNING_KEY),
                nulls_distinct=False,
                name="%(app_label)s_%(class)s_unique_signing_key_fingerprints",
            ),
            CheckConstraint(
                check=(
                    ~models.Q(category=AssetCategory.APT_AUTHENTICATION)
                    | models.Q(data__name__regex=r"^[A-Za-z][A-Za-z0-9+._-]*$")
                ),
                name="%(app_label)s_%(class)s_apt_auth_name",
            ),
            JsonDataUniqueConstraint(
                fields=["workspace", "data->>'name'"],
                condition=models.Q(category=AssetCategory.APT_AUTHENTICATION),
                nulls_distinct=False,
                name="%(app_label)s_%(class)s_unique_apt_auth_workspace_name",
            ),
            # Some categories of asset can have null workspaces, but not
            # signing keys.
            CheckConstraint(
                check=~models.Q(category=AssetCategory.SIGNING_KEY)
                | models.Q(workspace__isnull=False),
                name="%(app_label)s_%(class)s_workspace_not_null",
            ),
        ]

    @override
    def __str__(self) -> str:
        """Return basic information of Asset."""
        return (
            f"Id: {self.id} "
            f"Category: {self.category} "
            f"Workspace: {self.workspace}"
        )

    @override
    @copy_signature_from(models.Model.save)
    def save(self, **kwargs: Any) -> None:
        """Wrap save with permission checks."""
        from debusine.db.context import context

        if context.permission_checks_disabled:
            pass
        elif self._state.adding:
            enforce(self.can_create)
        else:
            enforce(self.can_edit)

        return super().save(**kwargs)

    @override
    def clean(self) -> None:
        """
        Ensure that data is valid for this asset category.

        :raise ValidationError: for invalid data.
        """
        self.data_model

    @property
    def slug(self) -> str:
        """Return a string slug that uniquely identifies the asset."""
        match self.category:
            case AssetCategory.SIGNING_KEY:
                data_model = self.data_model
                assert isinstance(data_model, SigningKeyData)
                return f"{data_model.purpose}:{data_model.fingerprint}"
            case AssetCategory.APT_AUTHENTICATION:
                data_model = self.data_model
                assert isinstance(data_model, BasicAPTAuthenticationData)
                if self.workspace is None:
                    return f":{data_model.name}"
                else:
                    return f"{self.workspace.name}:{data_model.name}"
            case _:
                raise NotImplementedError(
                    f"No slug defined for category '{self.category}'"
                )

    @permission_check("{user} cannot edit asset {resource}")
    def can_edit(self, pc: PermissionContext) -> bool:
        """Can user edit this asset."""
        assert pc.user.is_authenticated
        return self.roles.filter(
            group__users=pc.user, role=AssetRoles.OWNER
        ).exists()

    @permission_check(
        "{user} cannot create assets in {resource.workspace.scope}",
        work_request=Allow.PASS,
        anonymous=Allow.PASS,
    )
    def can_create(self, pc: PermissionContext) -> bool:
        """Can user create this asset."""
        # Allow signing workers to create Assets until we have delegated work
        # request permissions (#634)
        if pc.work_request:
            assert pc.work_request.worker is not None
            worker = pc.work_request.worker
            if worker.worker_type == WorkerType.SIGNING:
                return True
        if not pc.user.is_authenticated:
            return False
        if not self.workspace:
            return False
        if self.category == AssetCategory.CLOUD_PROVIDER_ACCOUNT:
            # Not currently creatable through the API
            return False
        return any(
            r.implies(Workspace.Roles.OWNER)
            for r in Workspace.Roles.from_iterable(
                self.workspace.scope.get_group_roles(pc)
            )
        )

    @property
    def data_model(self) -> BaseAssetDataModel:
        """Instantiate AssetData from data."""
        if not isinstance(self.data, dict):
            raise ValidationError({"data": "data must be a dictionary"})

        try:
            return asset_data_model(self.category, self.data)
        except ValueError as e:
            raise ValidationError(
                {
                    "category": (
                        f"{self.category}: invalid asset category or data: {e}"
                    ),
                },
            ) from e


class AssetRole(models.Model):
    """Role assignment for assets."""

    resource = models.ForeignKey(
        Asset,
        on_delete=models.CASCADE,
        related_name="roles",
    )

    group = models.ForeignKey(
        "Group",
        on_delete=models.CASCADE,
        related_name="asset_roles",
    )

    role = models.CharField(max_length=16, choices=AssetRoles.choices)


class AssetUsage(models.Model):
    """Usage of an Asset within a workspace."""

    Roles: TypeAlias = AssetUsageRoles

    asset = models.ForeignKey(
        Asset,
        on_delete=models.CASCADE,
        related_name="usage",
    )

    workspace = models.ForeignKey(
        "Workspace",
        on_delete=models.CASCADE,
        related_name="asset_usage",
    )

    objects = AssetUsageManager.from_queryset(AssetUsageQuerySet)()

    class Meta(TypedModelMeta):
        base_manager_name = "objects"
        constraints = [
            UniqueConstraint(
                fields=["asset", "workspace"],
                name="%(app_label)s_%(class)s_unique_asset_workspace",
            )
        ]

    def has_role(self, pc: PermissionContext, role: AssetUsageRoles) -> bool:
        """Check if the user has the given role on this AssetUsage."""
        return (
            AssetUsage.objects.with_role(pc, role).filter(pk=self.pk).exists()
        )

    @permission_check(
        "{user} cannot sign with {resource}", work_request=Allow.PASS
    )
    def can_sign_with(self, pc: PermissionContext) -> bool:
        """Check if the user can sign with this resource."""
        return self.has_role(pc, AssetUsageRoles.SIGNER)

    @permission_check(
        "{user} cannot sign a repository with {resource}",
        work_request=Allow.PASS,
    )
    def can_sign_repository_with(self, pc: PermissionContext) -> bool:
        """Check if the user can sign a repository with this resource."""
        return self.has_role(pc, AssetUsageRoles.REPOSITORY_SIGNER)

    @permission_check(
        "{user} cannot use {resource} for APT authentication",
        work_request=Allow.PASS,
    )
    def can_use_apt_authentication_with(self, pc: PermissionContext) -> bool:
        """Check if the user can use APT authentication with this resource."""
        return self.has_role(pc, AssetUsageRoles.APT_AUTHENTICATOR)


class AssetUsageRole(models.Model):
    """Role assignment for assets within a workspace."""

    resource = models.ForeignKey(
        AssetUsage,
        on_delete=models.CASCADE,
        related_name="roles",
    )

    group = models.ForeignKey(
        "Group",
        on_delete=models.CASCADE,
        related_name="asset_usage_roles",
    )

    role = models.CharField(max_length=32, choices=AssetUsageRoles.choices)


def get_public_keys(signing_keys: Iterable[Asset]) -> Generator[bytes]:
    """Yield public keys from each of some signing-key assets."""
    for signing_key in signing_keys:
        signing_key_data = signing_key.data_model
        assert isinstance(signing_key_data, SigningKeyData)
        yield signing_key_data.public_key



# Copyright © The Debusine Developers
# See the AUTHORS file at the top-level directory of this distribution
#
# This file is part of Debusine. It is subject to the license terms
# in the LICENSE file found in the top-level directory of this
# distribution. No part of Debusine, including this file, may be copied,
# modified, propagated, or distributed except according to the terms
# contained in the LICENSE file.

"""Tests for serializers."""

import base64
import datetime as dt
from typing import Any, ClassVar, override
from unittest import mock

from django.db import transaction
from django.db.models import Max
from django.test import RequestFactory, override_settings
from django.urls import reverse
from django.utils import timezone
from rest_framework.exceptions import ErrorDetail, ValidationError

from debusine.artifacts.models import (
    ArtifactCategory,
    BareDataCategory,
    CollectionCategory,
    DebusinePromise,
    DebusineTaskConfiguration,
    RuntimeStatistics,
)
from debusine.assets import AssetCategory, KeyPurpose, SigningKeyData
from debusine.client.models import LookupChildType, LookupResultType
from debusine.db.context import context
from debusine.db.models import (
    Artifact,
    Collection,
    CollectionItem,
    CollectionTokenGrant,
    File,
    FileInArtifact,
    Group,
    Token,
    WorkRequest,
    WorkflowTemplate,
)
from debusine.db.models.lookups import (
    LookupResultArtifact,
    LookupResultBare,
    LookupResultCollection,
)
from debusine.db.permissioncontext import PermissionContext
from debusine.db.playground import scenarios
from debusine.db.tests.utils import _calculate_hash_from_data
from debusine.server.serializers import (
    ArtifactSerializer,
    ArtifactSerializerResponse,
    AssetSerializer,
    CollectionSerializer,
    CollectionTokenGrantSerializer,
    CreateWorkflowRequestSerializer,
    FileSerializer,
    LookupMultipleSerializer,
    LookupResponseSerializer,
    LookupSingleSerializer,
    TaskConfigurationCollectionContentsSerializer,
    TaskConfigurationCollectionSerializer,
    WorkRequestCompletedSerializer,
    WorkRequestSerializer,
    WorkerRegisterSerializer,
    WorkflowTemplateSerializer,
    WorkspaceChainItemField,
    WorkspaceChainSerializer,
)
from debusine.tasks.models import OutputData, SbuildData, SbuildInput
from debusine.test import test_utils
from debusine.test.django import TestCase, TransactionTestCase


class WorkRequestSerializerTests(TestCase):
    """Tests for WorkRequestSerializer."""

    scenario = scenarios.DefaultContextAPI()

    work_request: ClassVar[WorkRequest]
    work_request_serializer: ClassVar[WorkRequestSerializer]
    source_artifact: ClassVar[Artifact]

    @override
    @classmethod
    def setUpTestData(cls) -> None:
        super().setUpTestData()

        cls.source_artifact = cls.playground.create_source_artifact()
        cls.playground.create_debian_environment(
            codename="sid", architecture="amd64"
        )

        worker = cls.playground.create_worker(fqdn="worker.lan")

        cls.work_request = cls.playground.create_worker_task(
            started_at=dt.datetime(2022, 2, 20, 15, 19, 1, 158424, dt.UTC),
            completed_at=dt.datetime(2022, 2, 20, 16, 2, 3, 558425, dt.UTC),
            worker=worker,
            task_name='sbuild',
            task_data=SbuildData(
                input=SbuildInput(source_artifact=cls.source_artifact.pk),
                build_architecture="amd64",
                environment="debian/match:codename=sid",
            ),
            created_by=cls.scenario.user,
            result=WorkRequest.Results.SUCCESS,
        )

        cls.work_request_serializer = WorkRequestSerializer(cls.work_request)

    def test_expected_fields(self) -> None:
        """Serializer returns the expected fields."""
        self.assertCountEqual(
            self.work_request_serializer.data.keys(),
            {
                'aborted_by',
                'artifacts',
                'completed_at',
                'completed_by',
                'created_at',
                'created_by',
                'duration',
                'dynamic_task_data',
                'event_reactions',
                'id',
                'links',
                'parent',
                'priority_adjustment',
                'priority_base',
                'result',
                'scheduler_tags_provided',
                'scheduler_tags_required',
                'scope',
                'started_at',
                'status',
                'task_data',
                'task_name',
                'task_type',
                'url',
                'worker',
                'workflow_root',
                'workflow_data',
                'workspace',
            },
        )

    def test_serialize_include_artifact_ids_created_by_work_request(
        self,
    ) -> None:
        """Assert serialized WorkRequest include artifact ids created by it."""
        artifact_1, _ = self.playground.create_artifact(
            work_request=self.work_request
        )
        artifact_2, _ = self.playground.create_artifact(
            work_request=self.work_request
        )
        serialized = self.work_request_serializer.data
        self.assertEqual(
            serialized["artifacts"], sorted([artifact_1.id, artifact_2.id])
        )

    def assert_is_valid(
        self,
        data: dict[str, Any],
        only_fields: list[str] | None,
        is_valid_expected: bool,
        errors: dict[str, Any] | None = None,
    ) -> WorkRequestSerializer:
        """Validate data using only_fields and expects is_valid_expected."""
        work_request_serializer = WorkRequestSerializer(
            data=data, only_fields=only_fields
        )
        self.assertEqual(work_request_serializer.is_valid(), is_valid_expected)

        if errors is not None:
            self.assertEqual(work_request_serializer.errors, errors)

        return work_request_serializer

    def test_validate_only_fields_invalid(self) -> None:
        """Use only_fields with an extra field."""
        self.scenario.set_current()
        data = {
            'task_name': 'sbuild',
            'task_data': {'foo': 'bar'},
            'unwanted_field': 'for testing',
            'unwanted_field2': 'for testing',
            'binnmu': {'changelog': 'not allowed here', 'suffix': '+b1'},
            'workspace': self.scenario.workspace.name,
            'created_by': self.scenario.user_token.user_id,
        }
        errors = {
            'non_field_errors': [
                ErrorDetail(
                    string='Invalid fields: binnmu,'
                    ' unwanted_field, unwanted_field2',
                    code='invalid',
                )
            ]
        }
        self.assert_is_valid(
            data,
            ['task_name', 'task_data', 'workspace', 'created_by'],
            False,
            errors,
        )

    def test_validate_only_fields_valid(self) -> None:
        """Use only_fields with an extra field in data."""
        self.scenario.set_current()
        data = {
            'task_name': 'sbuild',
            'task_data': {'foo': 'bar'},
            'workspace': self.scenario.workspace.name,
            'created_by': self.scenario.user_token.user_id,
        }
        self.assert_is_valid(
            data, ['task_name', 'task_data', 'workspace', 'created_by'], True
        )

    def test_validate_without_only_fields(self) -> None:
        """is_valid() return True: not using only_fields."""
        self.scenario.set_current()
        data = {
            'task_name': 'sbuild',
            'workspace': self.scenario.workspace.name,
            'created_by': self.scenario.user_token.user_id,
        }
        self.assert_is_valid(data, None, True)

    def test_serialized_scope_and_workspace_names(self) -> None:
        """Serialized work request has scope and workspace names."""
        work_request_serialized = WorkRequestSerializer(self.work_request).data

        self.assertEqual(
            work_request_serialized["scope"], self.scenario.scope.name
        )
        self.assertEqual(
            work_request_serialized["workspace"], self.scenario.workspace.name
        )

    def test_workspace_name_resolved_in_current_scope(self) -> None:
        """The given workspace name is resolved in the current scope."""
        scope1 = self.playground.get_or_create_scope("scope1")
        scope2 = self.playground.get_or_create_scope("scope2")
        scope3 = self.playground.get_or_create_scope("scope3")
        workspace1 = self.playground.create_workspace(
            scope=scope1, name="common-name", public=True
        )
        workspace2 = self.playground.create_workspace(
            scope=scope2, name="common-name", public=True
        )
        data = {
            "task_name": "sbuild",
            "workspace": "common-name",
            "created_by": self.scenario.user_token.user_id,
        }
        only_fields = ["task_name", "workspace", "created_by"]

        for workspace in (workspace1, workspace2):
            with context.local():
                context.set_permission_context_scoped(
                    workspace.scope, PermissionContext.for_ui_anonymous()
                )
                work_request_serializer = self.assert_is_valid(
                    data, only_fields, True
                )
                self.assertEqual(
                    work_request_serializer.validated_data["workspace"],
                    workspace,
                )

        with context.local():
            context.set_permission_context_scoped(
                scope3, PermissionContext.for_ui_anonymous()
            )
            self.assert_is_valid(data, only_fields, False)

    def test_default_workspace(self) -> None:
        """Use the default workspace name in the current scope, if it exists."""
        data = {
            "task_name": "noop",
            "created_by": self.scenario.user_token.user_id,
        }
        only_fields = ["task_name", "workspace", "created_by"]

        with context.local():
            self.scenario.set_current()
            work_request_serializer = self.assert_is_valid(
                data, only_fields, True
            )
            self.assertEqual(
                work_request_serializer.validated_data["workspace"],
                self.playground.get_default_workspace(),
            )

        with context.local():
            context.set_permission_context_scoped(
                self.playground.get_or_create_scope("empty-scope"),
                PermissionContext.for_ui_anonymous(),
            )
            self.assert_is_valid(
                data,
                only_fields,
                False,
                {
                    "workspace": [
                        ErrorDetail(
                            string="This field is required.", code="required"
                        )
                    ]
                },
            )

    def test_readonly_fields(self) -> None:
        self.scenario.set_current()
        work_request_serialized = WorkRequestSerializer(self.work_request).data
        for field, edited_value in (
            ("scope", "newscope"),
            ("scheduler_tags_provided", ["changed"]),
            ("scheduler_tags_required", ["changed"]),
        ):
            with self.subTest(field=field), context.local():
                data = work_request_serialized.copy()
                data[field] = edited_value
                serializer = WorkRequestSerializer(self.work_request, data=data)
                self.assertTrue(serializer.is_valid())
                self.assertEqual(serializer.errors, {})
                new_instance = serializer.save()
                assert (
                    WorkRequestSerializer(new_instance).data
                    == work_request_serialized
                )


class WorkerRegisterSerializerTests(TestCase):
    """Test for WorkerRegisterSerializer class."""

    def test_expected_fields(self) -> None:
        """Expected fields are defined in the serializer."""
        worker_register_serializer = WorkerRegisterSerializer()
        data = worker_register_serializer.data

        self.assertCountEqual(data.keys(), {'token', 'fqdn', 'worker_type'})


class WorkRequestCompletedSerializerTests(TestCase):
    """Test for WorkRequestCompletedSerializer class."""

    def test_accept_success(self) -> None:
        """Serializer accepts {"result": "success"}."""
        work_request_completed_serializer = WorkRequestCompletedSerializer(
            data={'result': 'success'}
        )
        self.assertTrue(work_request_completed_serializer.is_valid())
        self.assertEqual(
            work_request_completed_serializer.validated_data['result'],
            'success',
        )

    def test_accept_success_with_output_data(self) -> None:
        """Serializer accepts output data."""
        work_request_completed_serializer = WorkRequestCompletedSerializer(
            data={
                "result": "success",
                "output_data": OutputData(
                    runtime_statistics=RuntimeStatistics(duration=60)
                ).model_dump(mode="json", exclude_unset=True),
            }
        )
        self.assertTrue(work_request_completed_serializer.is_valid())
        self.assertEqual(
            work_request_completed_serializer.validated_data['result'],
            'success',
        )

    def test_not_accept_unknown_result(self) -> None:
        """Serializer does not accept unrecognised result."""
        work_request_completed_serializer = WorkRequestCompletedSerializer(
            data={'result': 'something'}
        )
        self.assertFalse(work_request_completed_serializer.is_valid())
        self.assertIn("result", work_request_completed_serializer.errors)

    def test_not_accept_invalid_output_data(self) -> None:
        """Serializer does not accept invalid output data."""
        work_request_completed_serializer = WorkRequestCompletedSerializer(
            data={"result": "success", "output_data": {"nonsense": True}}
        )
        self.assertFalse(work_request_completed_serializer.is_valid())
        self.assertIn("output_data", work_request_completed_serializer.errors)


def serialized_file() -> dict[str, Any]:
    """Return a file."""
    return {
        "type": "file",
        "size": 3827,
        "checksums": {
            "sha256": "164a3bc86c0fe9a7aa15cfa9156e9be1124aad69e"
            "a757a011be1f1a13502409f",
            "md5": "7d13cb5b2bee07003d2b69ccbd256e65",
        },
    }


class FileSerializerTests(TestCase):
    """Test for FileSerializer class."""

    def test_serializer_with_valid_data(self) -> None:
        """Test File serializer (valid data)."""
        data = serialized_file()

        file_serializer = FileSerializer(data=data)

        self.assertTrue(file_serializer.is_valid())
        self.assertEqual(file_serializer.validated_data, data)

    def test_is_valid_false_missing_file_type(self) -> None:
        """Test FileSerializer (invalid data, no file_type)."""
        data = serialized_file()

        del data["type"]

        file_serializer = FileSerializer(data=data)

        self.assertFalse(file_serializer.is_valid())

    def test_is_valid_false_invalid_file_type(self) -> None:
        """Test FileSerializer (invalid file type)."""
        file = serialized_file()

        file["type"] = "directory"

        file_serializer = FileSerializer(data=file)

        self.assertFalse(file_serializer.is_valid())

    def test_is_valid_false_invalid_file_size(self) -> None:
        """Test FileSerializer (size must be positive)."""
        file = serialized_file()

        file["size"] = -10

        file_serializer = FileSerializer(data=file)

        self.assertFalse(file_serializer.is_valid())


class FileSerializerResponseTests(TestCase):
    """Tests for FileSerializerResponse."""

    def test_url(self) -> None:
        """It has the URL of the file."""
        file = serialized_file()
        file["url"] = "https://example.com/some/URL"

        FileSerializer(data=file)


class ArtifactSerializerTests(TestCase):
    """Test for ArtifactSerializer class."""

    scenario = scenarios.DefaultContext(set_current=True)

    def test_is_valid_true(self) -> None:
        """Test Artifact serializer (valid data)."""
        category = "artifact-test"

        serialized_artifact = {
            "category": category,
            "workspace": self.scenario.workspace.name,
            "work_request": 5,
            "files": {"AUTHORS": serialized_file()},
            "data": {"key1": "value1", "key2": "value2"},
        }

        artifact_serializer = ArtifactSerializer(data=serialized_artifact)
        self.assertTrue(artifact_serializer.is_valid())
        self.assertEqual(
            artifact_serializer.validated_data,
            {**serialized_artifact, "workspace": self.scenario.workspace},
        )

    def test_is_valid_true_no_work_request(self) -> None:
        """Test Artifact serializer (valid data) with no work_request."""
        category = "artifact-test"

        serialized_artifact = {
            "category": category,
            "workspace": self.scenario.workspace.name,
            "files": {"AUTHORS": serialized_file()},
            "data": {"key1": "value1", "key2": "value2"},
        }
        artifact_serializer = ArtifactSerializer(data=serialized_artifact)
        self.assertTrue(artifact_serializer.is_valid())
        self.assertNotIn("work_request", artifact_serializer.validated_data)

    def test_is_valid_true_work_request_is_none(self) -> None:
        """Test Artifact serializer (valid data) with work_request=None."""
        category = "artifact-test"

        serialized_artifact = {
            "category": category,
            "workspace": self.scenario.workspace.name,
            "files": {"AUTHORS": serialized_file()},
            "data": {"key1": "value1", "key2": "value2"},
            "work_request": None,
        }
        artifact_serializer = ArtifactSerializer(data=serialized_artifact)
        self.assertTrue(artifact_serializer.is_valid())
        self.assertIsNone(artifact_serializer.validated_data["work_request"])


class ArtifactSerializerResponseTests(TestCase):
    """Tests for ArtifactSerializerResponse class."""

    files_to_add: list[str]
    artifact: ClassVar[Artifact]
    files: ClassVar[dict[str, bytes]]

    @override
    @classmethod
    def setUpTestData(cls) -> None:
        super().setUpTestData()
        cls.files_to_add = ["Makefile", "README.txt"]
        cls.artifact, cls.files = cls.playground.create_artifact(
            cls.files_to_add, create_files=True, skip_add_files_in_store=True
        )

    @classmethod
    def get_links(cls, artifact: Artifact) -> dict[str, str | dict[str, str]]:
        return {
            "self": reverse(
                "api:artifact",
                kwargs={"artifact_id": artifact.pk},
            ),
            "webui_self": artifact.get_absolute_url(),
            "download_tar_gz": {
                "href": artifact.get_absolute_url_download()
                + "?archive=tar.gz",
                "type": "application/tar+gzip",
            },
        }

    @override_settings(ALLOWED_HOSTS=["*"])
    def test_from_artifact(self) -> None:
        """from_artifact() return the expected ArtifactSerializerResponse."""
        host = "example.com"
        request = RequestFactory().get("/test", HTTP_HOST=host)
        serializer = ArtifactSerializerResponse.from_artifact(
            self.artifact, request
        )

        download_path = self.artifact.get_absolute_url_download()
        download_url = f"http://{host}{download_path}?archive=tar.gz"
        created_at = test_utils.date_time_to_isoformat_rest_framework(
            self.artifact.created_at
        )

        files = {}

        for file_name, file_content in self.files.items():
            if file_name == "README.txt":
                content_type = "text/plain; charset=utf-8"
                file_in_artifact = self.artifact.fileinartifact_set.get(
                    path=file_name
                )
                file_in_artifact.content_type = content_type
                file_in_artifact.save()
            else:
                content_type = None

            checksums = {
                File.current_hash_algorithm: _calculate_hash_from_data(
                    file_content
                ).hex()
            }
            path = reverse(
                "workspaces:artifacts:file-download",
                kwargs={
                    "wname": self.artifact.workspace.name,
                    "artifact_id": self.artifact.id,
                    "path": file_name,
                },
            )
            files[file_name] = {
                "size": len(file_content),
                "checksums": checksums,
                "type": "file",
                "url": f"http://{host}{path}",
                "content_type": content_type,
            }

        assert serializer.data == {
            "id": self.artifact.id,
            "url": self.artifact.get_absolute_url(),
            "scope": self.artifact.workspace.scope.name,
            "workspace": self.artifact.workspace.name,
            "category": str(self.artifact.category),
            "data": self.artifact.data,
            "created_at": created_at,
            "expire_at": self.artifact.expire_at,
            "download_tar_gz_url": download_url,
            "files_to_upload": self.files_to_add,
            "files": files,
            "links": self.get_links(self.artifact),
        }

    @override_settings(ALLOWED_HOSTS=["*"])
    def test_build_absolute_download_url(self) -> None:
        """_download_url() return the expected URL."""
        host = "example.com"

        request = RequestFactory().get("/test", HTTP_HOST=host)

        actual = ArtifactSerializerResponse._build_absolute_download_url(
            self.artifact, request
        )

        path = self.artifact.get_absolute_url_download()
        expected = f"{request.scheme}://{host}{path}"

        self.assertEqual(actual, expected)


class ArtifactSerializerResponseTransactionTests(TransactionTestCase):
    """Tests for ArtifactSerializerResponse class that require transactions."""

    @override_settings(ALLOWED_HOSTS=["*"])
    def test_files_to_upload_race(self) -> None:
        """
        ``files_to_upload`` ignores files created in other transactions.

        Since Debusine currently uses the ``READ COMMITTED`` transaction
        isolation level, it can experience "phantom reads", where a
        transaction re-executes a query and finds that the set of rows
        satisfying its search condition has changed due to another
        recently-committed transaction.  The serializer does not get
        confused by this when computing ``files_to_upload``, even if another
        transaction races with it.
        """
        with transaction.atomic():
            # Ensure that the playground's default workspace has been
            # created with an ID visible to both transactions.
            self.playground.get_default_workspace()

            # Create a pre-existing file so that we can make sure that
            # different transactions reference the same file.  It isn't in
            # any artifact, so artifacts will be considered as needing to
            # upload it.
            common_file_contents = b"Common file contents"
            common_file = self.playground.create_file_in_backend(
                self.playground.get_default_file_store().get_backend_object(),
                common_file_contents,
            )

        with transaction.atomic():
            artifact, _ = self.playground.create_artifact(
                {"common-file": common_file_contents},
                create_files=True,
                skip_add_files_in_store=True,
            )

            with self.separate_connection(racing_alias := "racing"):
                with (
                    transaction.atomic(using=racing_alias),
                    context.disable_permission_checks(),
                ):
                    racing_artifact = Artifact.objects.using(
                        racing_alias
                    ).create(
                        category=artifact.category,
                        workspace=artifact.workspace,
                        data=artifact.data,
                    )
                    FileInArtifact.objects.using(racing_alias).create(
                        artifact=racing_artifact,
                        file_id=common_file.id,
                        path="common-file",
                        complete=True,
                    )

                host = "example.com"
                request = RequestFactory().get("/test", HTTP_HOST=host)

                serializer = ArtifactSerializerResponse.from_artifact(
                    artifact, request
                )

                self.assertEqual(
                    serializer.data["files_to_upload"], ["common-file"]
                )


class AssetSerializerTests(TestCase):
    """Tests for AssetSerializer."""

    scenario = scenarios.DefaultContext(set_current=True)

    def test_unknown_asset_category(self) -> None:
        """Deserialize an asset of an unknown category."""
        serializer = AssetSerializer(
            data={
                "category": "unknown",
                "workspace": self.scenario.workspace.name,
                "data": {},
                "work_request": 12,
            },
        )
        self.assertFalse(serializer.is_valid())
        self.assertIn("category", serializer.errors)
        self.assertEqual(
            serializer.errors["category"][0], "unknown is not a known category."
        )

    def test_invalid_data(self) -> None:
        """Deserialize an asset with invalid data."""
        serializer = AssetSerializer(
            data={
                "category": AssetCategory.SIGNING_KEY,
                "workspace": self.scenario.workspace.name,
                "data": {},
                "work_request": 12,
            },
        )
        self.assertFalse(serializer.is_valid())
        self.assertIn("data", serializer.errors)
        self.assertTrue(
            serializer.errors["data"][0].startswith("invalid asset data:")
        )

    def test_valid_data_work_request(self) -> None:
        """Deserialize an asset with valid data, created by a work request."""
        work_request = self.playground.create_worker_task()
        serializer = AssetSerializer(
            data={
                "category": AssetCategory.SIGNING_KEY,
                "workspace": self.scenario.workspace.name,
                "data": SigningKeyData(
                    description="A Test Key",
                    fingerprint="ABC123",
                    public_key=base64.b64encode(b"PUBLIC KEY").decode(),
                    purpose=KeyPurpose.OPENPGP_UPLOAD,
                ).model_dump(),
                "work_request": work_request.id,
            },
        )
        self.assertTrue(serializer.is_valid())
        with context.disable_permission_checks():
            asset = serializer.create(serializer.validated_data)
        self.assertEqual(asset.created_by_work_request, work_request)
        self.assertIsNone(asset.created_by)

    def test_valid_data_user(self) -> None:
        """Deserialize an asset with valid data, created by a user."""
        serializer = AssetSerializer(
            data={
                "category": AssetCategory.SIGNING_KEY,
                "workspace": self.scenario.workspace.name,
                "data": SigningKeyData(
                    description="A Test Key",
                    fingerprint="ABC123",
                    public_key=base64.b64encode(b"PUBLIC KEY").decode(),
                    purpose=KeyPurpose.OPENPGP_UPLOAD,
                ).model_dump(),
            },
        )
        self.assertTrue(serializer.is_valid())
        with context.disable_permission_checks():
            asset = serializer.create(serializer.validated_data)
        self.assertIsNone(asset.created_by_work_request)
        self.assertEqual(asset.created_by, self.scenario.user)


class LookupSingleSerializerTests(TestCase):
    """Tests for LookupSingleSerializer."""

    scenario = scenarios.DefaultContext(set_current=True)

    def test_expect_type(self) -> None:
        """Test possible values of `expect_type`."""
        for expect_type, validated_expect_type in (
            # All the items in LookupChildType are valid.
            *((str(item), item) for item in LookupChildType),
            # These specific values are mapped to their corresponding items
            # in LookupChildType, for compatibility with old clients.
            ("b", LookupChildType.BARE),
            ("a", LookupChildType.ARTIFACT),
            ("c", LookupChildType.COLLECTION),
        ):
            with self.subTest(expect_type=expect_type):
                data = {
                    "lookup": {"collection": "test"},
                    "workspace": self.playground.get_default_workspace().name,
                    "expect_type": expect_type,
                }

                serializer = LookupSingleSerializer(data=data)

                self.assertTrue(serializer.is_valid())
                self.assertEqual(
                    serializer.validated_data["expect_type"],
                    validated_expect_type,
                )


class LookupMultipleSerializerTests(TestCase):
    """Tests for LookupMultipleSerializer."""

    scenario = scenarios.DefaultContext(set_current=True)

    def test_expect_type(self) -> None:
        """Test possible values of `expect_type`."""
        for expect_type, validated_expect_type in (
            # All the items in LookupChildType are valid.
            *((str(item), item) for item in LookupChildType),
            # These specific values are mapped to their corresponding items
            # in LookupChildType, for compatibility with old clients.
            ("b", LookupChildType.BARE),
            ("a", LookupChildType.ARTIFACT),
            ("c", LookupChildType.COLLECTION),
        ):
            with self.subTest(expect_type=expect_type):
                data = {
                    "lookup": "x",
                    "workspace": self.playground.get_default_workspace().name,
                    "expect_type": expect_type,
                }

                serializer = LookupMultipleSerializer(data=data)

                self.assertTrue(serializer.is_valid())
                self.assertEqual(
                    serializer.validated_data["expect_type"],
                    validated_expect_type,
                )


class LookupResponseSerializerTests(TestCase):
    """Tests for LookupResponseSerializer."""

    def test_serialize_bare(self) -> None:
        collection = self.playground.create_collection(
            name="test", category=CollectionCategory.TEST
        )
        item = self.playground.create_bare_data_item(
            collection,
            "build-all",
            category=BareDataCategory.PROMISE,
            data=DebusinePromise(
                promise_work_request_id=11,
                promise_workflow_id=10,
                promise_category=ArtifactCategory.UPLOAD,
            ),
        )

        data = LookupResponseSerializer().to_representation(
            LookupResultBare(
                lookup="something",
                parent_collection_lookup=str(collection),
                collection_item=item,
            )
        )
        self.assertEqual(
            data,
            {
                "artifact": None,
                "collection": None,
                "collection_item": {
                    "name": "build-all",
                    "data": item.data,
                },
                "links": {
                    "parent_collection": reverse(
                        "api:collection-detail",
                        kwargs={
                            "pk": collection.pk,
                            "workspace": collection.workspace.name,
                        },
                    ),
                    "webui_parent_collection": collection.get_absolute_url(),
                },
                "parent_collection_lookup": str(collection),
                "result_type": LookupResultType.BARE,
            },
        )

    def test_serialize_artifact(self) -> None:
        collection = self.playground.create_collection(
            name="test", category=CollectionCategory.SUITE
        )
        srcpkg = self.playground.create_minimal_source_package_artifact()
        item = collection.manager.add_artifact(
            srcpkg,
            user=self.playground.get_default_user(),
            variables={"component": "main", "section": "games"},
        )

        data = LookupResponseSerializer().to_representation(
            LookupResultArtifact(
                parent_collection_lookup=str(collection),
                lookup="something",
                collection_item=item,
                artifact=srcpkg,
            )
        )
        self.assertEqual(
            data,
            {
                "artifact": srcpkg.pk,
                "collection": None,
                "collection_item": {
                    "name": item.name,
                    "data": item.data,
                },
                "parent_collection_lookup": str(collection),
                "links": {
                    "self": reverse(
                        "api:artifact", kwargs={"artifact_id": srcpkg.pk}
                    ),
                    "webui_self": srcpkg.get_absolute_url(),
                    "parent_collection": reverse(
                        "api:collection-detail",
                        kwargs={
                            "pk": collection.pk,
                            "workspace": collection.workspace.name,
                        },
                    ),
                    "webui_parent_collection": collection.get_absolute_url(),
                },
                "result_type": LookupResultType.ARTIFACT,
            },
        )

    def test_serialize_collection(self) -> None:
        parent = self.playground.create_singleton_collection(
            category=CollectionCategory.ARCHIVE
        )
        child = self.playground.create_collection(
            name="child", category=CollectionCategory.SUITE
        )
        item = parent.manager.add_collection(
            child,
            user=self.playground.get_default_user(),
        )

        data = LookupResponseSerializer().to_representation(
            LookupResultCollection(
                lookup="something",
                parent_collection_lookup=str(parent),
                collection_item=item,
                collection=child,
            )
        )
        self.assertEqual(
            data,
            {
                "artifact": None,
                "collection": child.pk,
                "collection_item": {
                    "name": item.name,
                    "data": item.data,
                },
                "parent_collection_lookup": str(parent),
                "links": {
                    "self": reverse(
                        "api:collection-detail",
                        kwargs={
                            "pk": child.pk,
                            "workspace": child.workspace.name,
                        },
                    ),
                    "webui_self": child.get_absolute_url(),
                    "parent_collection": reverse(
                        "api:collection-detail",
                        kwargs={
                            "pk": parent.pk,
                            "workspace": parent.workspace.name,
                        },
                    ),
                    "webui_parent_collection": parent.get_absolute_url(),
                },
                "result_type": LookupResultType.COLLECTION,
            },
        )


class WorkflowTemplateSerializerTests(TestCase):
    """Tests for WorkflowTemplateSerializer."""

    scenario = scenarios.DefaultContextAPI(set_current=True)

    def test_parameters_order(self) -> None:
        template = WorkflowTemplate(
            name="test",
            workspace=self.scenario.workspace,
            task_name="debian_pipeline",
            static_parameters={
                "codename": "sid",
                "vendor": "debian",
                "sbuild_backend": "unshare",
            },
            runtime_parameters={},
            priority=0,
        )

        serializer = WorkflowTemplateSerializer()
        res = serializer.to_representation(template)

        sorted_params = list(res["static_parameters"].keys())
        expected_params = ["vendor", "codename", "sbuild_backend"]

        self.assertEqual(sorted_params, expected_params)

    def test_to_representation_invalid_task_name(self) -> None:
        template = WorkflowTemplate(
            name="test",
            workspace=self.scenario.workspace,
            task_name="testing",
            static_parameters={"vendor": "debian", "codename": "sid"},
            runtime_parameters={},
            priority=0,
        )
        serializer = WorkflowTemplateSerializer()
        res = serializer.to_representation(template)

        self.assertEqual(
            res["static_parameters"], {"vendor": "debian", "codename": "sid"}
        )

    def test_non_dict_params(self) -> None:
        template = WorkflowTemplate(
            name="test-list",
            workspace=self.scenario.workspace,
            task_name="debian_pipeline",
            static_parameters=["vendor", "codename", "suite"],
            runtime_parameters=None,
            priority=0,
        )
        serializer = WorkflowTemplateSerializer()
        res = serializer.to_representation(template)

        self.assertEqual(
            res["static_parameters"], ["vendor", "codename", "suite"]
        )

    def test_extra_parameters(self) -> None:
        template = WorkflowTemplate(
            name="test-extra",
            workspace=self.scenario.workspace,
            task_name="debian_pipeline",
            static_parameters={
                "version": "1.5.1",
                "vendor": "debian",
            },
            runtime_parameters={},
            priority=0,
        )
        serializer = WorkflowTemplateSerializer()
        res = serializer.to_representation(template)

        fields = list(res["static_parameters"].keys())
        self.assertEqual(fields, ["vendor", "version"])

    def test_valid_data(self) -> None:
        """Test valid data."""
        data = {
            "name": "wt",
            "task_name": "noop",
        }

        serializer = WorkflowTemplateSerializer(data=data)

        self.assertTrue(serializer.is_valid())
        self.assertEqual(serializer.validated_data, data)

    def test_invalid_task_name(self) -> None:
        """The serializer rejects an invalid task name."""
        data = {
            "name": "wt",
            "task_name": "nonexistent",
        }
        serializer = WorkflowTemplateSerializer(data=data)
        self.assertFalse(serializer.is_valid())

    def test_extra_groups(self) -> None:
        group1 = self.playground.create_group("group1")
        group2 = self.playground.create_group("group2")
        template = WorkflowTemplate.objects.create(
            name="test",
            workspace=self.scenario.workspace,
            task_name="noop",
        )
        template.extra_groups.set((group1, group2))

        serializer = WorkflowTemplateSerializer()
        res = serializer.to_representation(template)
        self.assertEqual(res["extra_groups"], [str(group1), str(group2)])

        # Also try resolving by ID
        res["extra_groups"][0] = group1.id
        deserializer = WorkflowTemplateSerializer(data=res)
        self.assertTrue(deserializer.is_valid())
        self.assertEqual(
            deserializer.validated_data["extra_groups"], [group1, group2]
        )

    def test_extra_groups_invalid_group_name(self) -> None:
        invalid_gid = (Group.objects.aggregate(Max("id"))['id__max'] or 0) + 1

        base_data = {"id": 1, "name": "noop", "task_name": "noop"}

        for extra_groups, error in (
            (
                [invalid_gid],
                ErrorDetail(
                    string=f"Group {invalid_gid} does not exist.",
                    code="does_not_exist",
                ),
            ),
            (
                [3.14],
                ErrorDetail(string="Invalid value.", code="invalid"),
            ),
            (
                ["does-not-exist"],
                ErrorDetail(string="Invalid value.", code="invalid"),
            ),
            (
                ["debusine/does-not-exist"],
                ErrorDetail(
                    string="Group debusine/does-not-exist does not exist.",
                    code="does_not_exist",
                ),
            ),
        ):
            with self.subTest(extra_groups=repr(extra_groups)):
                deserializer = WorkflowTemplateSerializer(
                    data={**base_data, "extra_groups": extra_groups}
                )
                self.assertFalse(deserializer.is_valid())
                self.assertIn("extra_groups", deserializer.errors)
                self.assertEqual(deserializer.errors["extra_groups"][0], error)


class CreateWorkflowRequestSerializerTests(TestCase):
    """Tests for :py:class:`CreateWorkflowRequestSerializer`."""

    scenario = scenarios.DefaultContextAPI(set_current=True)

    def test_validate_names(self) -> None:
        for data, expected_ok in (
            ({}, False),
            ({"template_name": "noop-template"}, True),
            ({"task_name": "noop"}, True),
            ({"template_name": "noop-template", "task_name": "noop"}, False),
        ):
            with self.subTest(data=str(data)):
                serializer = CreateWorkflowRequestSerializer(data=data)
                if expected_ok:
                    self.assertTrue(serializer.is_valid())
                else:
                    self.assertFalse(serializer.is_valid())
                    self.assertEqual(
                        serializer.errors,
                        {
                            "non_field_errors": [
                                ErrorDetail(
                                    string=(
                                        "Exactly one of template_name and"
                                        " task_name must be set"
                                    ),
                                    code="invalid",
                                )
                            ]
                        },
                    )


class TaskConfigurationCollectionContentsSerializerTests(TestCase):
    """Tests for :py:class:`TaskConfigurationCollectionContentsSerializer`."""

    scenario = scenarios.DefaultContextAPI(set_current=True)
    collection: ClassVar[Collection]

    @override
    @classmethod
    def setUpClass(cls) -> None:
        super().setUpClass()
        cls.collection = cls.scenario.default_task_configuration_collection

    def data_for_items(
        self, *items: DebusineTaskConfiguration
    ) -> dict[str, Any]:
        """Prepare serialized data for the given items."""
        return {
            "collection": TaskConfigurationCollectionSerializer(
                self.collection
            ).data,
            "items": [i.model_dump() for i in items],
        }

    def test_deserialize_empty(self) -> None:
        data = self.data_for_items()
        serializer = TaskConfigurationCollectionContentsSerializer(data=data)
        serializer.is_valid()
        self.assertEqual(
            serializer.validated_data["collection"],
            {
                "id": self.collection.id,
                "name": self.collection.name,
                "data": self.collection.data,
            },
        )
        self.assertEqual(serializer.validated_data["items"], [])

    def test_deserialize_items(self) -> None:
        sample1 = DebusineTaskConfiguration(template="sample1")
        sample2 = DebusineTaskConfiguration(template="sample2")
        data = self.data_for_items(sample1, sample2)
        serializer = TaskConfigurationCollectionContentsSerializer(data=data)
        serializer.is_valid()
        self.assertEqual(serializer.validated_data["items"], [sample1, sample2])

    def test_deserialize_items_name_conflict(self) -> None:
        sample1 = DebusineTaskConfiguration(template="sample1")
        data = self.data_for_items(sample1, sample1)
        serializer = TaskConfigurationCollectionContentsSerializer(data=data)
        self.assertFalse(serializer.is_valid())
        self.assertEqual(
            serializer.errors,
            {
                "items": [
                    ErrorDetail(
                        string="'template:sample1': found multiple items"
                        " with the same name",
                        code='invalid',
                    )
                ]
            },
        )


class WorkspaceChainItemFieldTests(TestCase):
    """Tests for :py:class:`WorkspaceChainItemField`."""

    scenario = scenarios.DefaultContextAPI(set_current=True)

    def test_to_representation(self) -> None:
        field = WorkspaceChainItemField()
        self.assertEqual(
            field.to_representation(self.scenario.workspace),
            {
                "id": self.scenario.workspace.pk,
                "scope": self.scenario.scope.name,
                "workspace": self.scenario.workspace.name,
            },
        )

    def test_from_id(self) -> None:
        field = WorkspaceChainItemField()
        self.assertEqual(
            field.to_internal_value({"id": self.scenario.workspace.pk}),
            self.scenario.workspace,
        )

    def test_from_id_cannot_display(self) -> None:
        field = WorkspaceChainItemField()
        hidden = self.playground.create_workspace(name="hidden")
        with self.assertRaisesRegex(
            ValidationError, r"Workspace with id \d+ does not exist"
        ):
            field.to_internal_value({"id": hidden.pk})

    def test_from_scope_workspace(self) -> None:
        field = WorkspaceChainItemField()
        self.assertEqual(
            field.to_internal_value(
                {
                    "scope": self.scenario.scope.name,
                    "workspace": self.scenario.workspace.name,
                }
            ),
            self.scenario.workspace,
        )

    def test_from_workspace_only(self) -> None:
        field = WorkspaceChainItemField()
        self.assertEqual(
            field.to_internal_value(
                {"workspace": self.scenario.workspace.name}
            ),
            self.scenario.workspace,
        )

    def test_from_mismatch_id_worksapce(self) -> None:
        field = WorkspaceChainItemField()
        with self.assertRaisesRegex(
            ValidationError, r"Workspace with id \d+ is not named 'invalid'"
        ):
            field.to_internal_value(
                {"id": self.scenario.workspace.id, "workspace": "invalid"}
            )
        with self.assertRaisesRegex(
            ValidationError, r"Workspace with id \d+ is not in scope 'invalid'"
        ):
            field.to_internal_value(
                {"id": self.scenario.workspace.id, "scope": "invalid"}
            )

    def test_from_str_cannot_display(self) -> None:
        field = WorkspaceChainItemField()
        self.playground.create_workspace(name="hidden")
        with self.assertRaisesRegex(
            ValidationError,
            r"Workspace 'hidden' does not exist in scope 'debusine'",
        ):
            field.to_internal_value({"workspace": "hidden"})

    def test_from_invalid(self) -> None:
        field = WorkspaceChainItemField()
        value: Any
        for value in (None, [], 3.14, False, True, "invalid", 42):
            with (
                self.subTest(value=value),
                self.assertRaisesRegex(
                    ValidationError,
                    r"Workspace indicator should be a dict",
                ),
            ):
                field.to_internal_value(value)

    def test_from_empty_dict(self) -> None:
        field = WorkspaceChainItemField()
        with self.assertRaisesRegex(
            ValidationError,
            r"at least id or workspace need to be set",
        ):
            field.to_internal_value({})


class WorkspaceChainSerializerTests(TestCase):
    """Tests for :py:class:`WorkspaceChainSerializer`."""

    scenario = scenarios.DefaultContextAPI(set_current=True)

    def test_valid_data(self) -> None:
        first = self.playground.create_workspace(name="first", public=True)
        second = self.playground.create_workspace(name="second", public=True)

        for data, expected in (
            ([], []),
            ([{"id": first.pk}, {"id": second.pk}], [first, second]),
            ([{"workspace": "first"}], [first]),
            ([{"scope": "debusine", "workspace": "first"}], [first]),
            (
                [
                    {"id": second.pk},
                    {"scope": "debusine", "workspace": "first"},
                ],
                [second, first],
            ),
        ):
            with self.subTest(data=data):
                serializer = WorkspaceChainSerializer(data={"chain": data})
                self.assertTrue(serializer.is_valid(raise_exception=True))
                self.assertEqual(serializer.validated_data, {"chain": expected})


class CollectionSerializerTests(TestCase):
    """Tests for :py:class:`CollectionSerializer`."""

    scenario = scenarios.DefaultContextAPI(set_current=True)

    sid: ClassVar[Collection]

    @override
    @classmethod
    def setUpTestData(cls) -> None:
        super().setUpTestData()
        cls.sid = cls.playground.create_collection(
            name="sid", category=CollectionCategory.SUITE
        )

    def create(self, **data: Any) -> CollectionSerializer:
        """Create a CollectionSerializer with data overrides."""
        defaults = CollectionSerializer(self.sid).data
        return CollectionSerializer(data=defaults | data)

    def test_name_validation(self) -> None:
        for value, expected in (
            ("sid", True),
            ("_", True),
            ("sid/test", False),
            ("_a", False),
            ("9000", False),
        ):
            with self.subTest(name=value):
                s = self.create(name=value)
                self.assertIs(s.is_valid(), expected)

    def test_category_validation(self) -> None:
        for value, expected in (
            ("debian:suite", True),
            ("foobar", False),
            ("9000", False),
        ):
            with self.subTest(category=value):
                s = self.create(category=value)
                self.assertIs(s.is_valid(), expected)

    def test_duration(self) -> None:
        for value, expected in (
            (None, None),
            (dt.timedelta(microseconds=1), 1),
            (dt.timedelta(milliseconds=1), 1),
            (dt.timedelta(days=0), 0),
            (dt.timedelta(days=1), 1),
            (dt.timedelta(hours=1), 1),
            (dt.timedelta(hours=25), 2),
            (dt.timedelta(seconds=1), 1),
            (dt.timedelta(days=3, seconds=1), 4),
            (dt.timedelta(days=365, hours=24), 366),
        ):
            with self.subTest(value=str(value)):
                # Test serialization
                self.sid.full_history_retention_period = value
                self.sid.metadata_only_retention_period = value
                serializer = CollectionSerializer(self.sid)
                self.assertEqual(
                    serializer.data["full_history_retention_period"], expected
                )
                self.assertEqual(
                    serializer.data["metadata_only_retention_period"], expected
                )

                # Test deserialization
                serializer = CollectionSerializer(
                    data={
                        "id": self.sid.id,
                        "name": self.sid.name,
                        "category": self.sid.category,
                        "full_history_retention_period": expected,
                        "metadata_only_retention_period": expected,
                        "data": {},
                    }
                )
                self.assertTrue(serializer.is_valid(raise_exception=True))
                self.assertEqual(
                    serializer.validated_data["full_history_retention_period"],
                    (
                        dt.timedelta(days=expected)
                        if expected is not None
                        else None
                    ),
                )
                self.assertEqual(
                    serializer.validated_data["metadata_only_retention_period"],
                    (
                        dt.timedelta(days=expected)
                        if expected is not None
                        else None
                    ),
                )

    def test_get_stats_disabled_in_context(self) -> None:
        serializer = CollectionSerializer(self.sid)
        with mock.patch.object(
            self.sid,
            "get_item_stats",
            return_value={
                ((CollectionItem.Types.BARE, True), 4),
                ((CollectionItem.Types.BARE, False), 5),
            },
        ):
            assert serializer.get_stats(self.sid) is None

    def test_get_stats(self) -> None:
        serializer = CollectionSerializer(self.sid)
        serializer.context["with_stats"] = True
        with mock.patch.object(
            self.sid,
            "get_item_stats",
            return_value=[
                {
                    "type": CollectionItem.Types.BARE,
                    "category": "category",
                    "active": True,
                    "count": 4,
                },
                {
                    "type": CollectionItem.Types.BARE,
                    "category": "category",
                    "active": False,
                    "count": 5,
                },
            ],
        ):
            assert serializer.get_stats(self.sid) == [
                {
                    "type": "bare",
                    "category": "category",
                    "active": True,
                    "count": 4,
                },
                {
                    "type": "bare",
                    "category": "category",
                    "active": False,
                    "count": 5,
                },
            ]


class CollectionTokenGrantSerializerTests(TestCase):
    """Tests for :py:class:`CollectionTokenGrantSerializer`."""

    scenario = scenarios.DefaultContextAPI(set_current=True)

    sid: ClassVar[Collection]
    token: ClassVar[Token]

    @override
    @classmethod
    def setUpTestData(cls) -> None:
        super().setUpTestData()
        cls.sid = cls.playground.create_collection(
            name="sid", category=CollectionCategory.SUITE
        )
        cls.token = cls.playground.create_external_user_token(username="test")

    def assertGrant(self, grant: CollectionTokenGrant) -> None:
        self.assertEqual(grant.collection, self.sid)
        self.assertEqual(grant.role, Collection.Roles.VIEWER)
        self.assertLessEqual(grant.created_at, timezone.now())
        self.assertEqual(grant.created_by, self.scenario.user)

    def test_valid_data(self) -> None:
        for data in (
            {"token_hash": self.token.hash, "role": Collection.Roles.VIEWER},
            {
                "username": "test",
                "comment": "a comment",
                "role": Collection.Roles.VIEWER,
            },
        ):
            with self.subTest(data=repr(data)):
                serializer = CollectionTokenGrantSerializer(data=data)
                self.assertTrue(serializer.is_valid(raise_exception=True))

    def test_no_token_hash_or_username(self) -> None:
        serializer = CollectionTokenGrantSerializer(
            data={"role": Collection.Roles.VIEWER}
        )
        self.assertFalse(serializer.is_valid())
        self.assertEqual(
            serializer.errors,
            {
                "non_field_errors": [
                    ErrorDetail(
                        string="Must provide either token_hash or username",
                        code="invalid",
                    )
                ]
            },
        )

    def test_save_from_token_hash(self) -> None:
        serializer = CollectionTokenGrantSerializer(
            data={
                "token_hash": self.token.hash,
                "role": Collection.Roles.VIEWER,
            }
        )
        self.assertTrue(serializer.is_valid())

        grant = serializer.save(
            collection=self.sid, created_by=self.scenario.user
        )

        self.assertGrant(grant)
        self.assertEqual(grant.token, self.token)

    def test_save_from_token_hash_and_matching_username(self) -> None:
        serializer = CollectionTokenGrantSerializer(
            data={
                "token_hash": self.token.hash,
                "username": self.token.external_username,
                "role": Collection.Roles.VIEWER,
            }
        )
        self.assertTrue(serializer.is_valid())

        grant = serializer.save(
            collection=self.sid, created_by=self.scenario.user
        )

        self.assertGrant(grant)
        self.assertEqual(grant.token, self.token)

    def test_save_from_token_hash_and_wrong_username(self) -> None:
        serializer = CollectionTokenGrantSerializer(
            data={
                "token_hash": self.token.hash,
                "username": "wrong",
                "role": Collection.Roles.VIEWER,
            }
        )
        self.assertTrue(serializer.is_valid())

        with self.assertRaisesRegex(
            ValidationError,
            f"Token {self.token} has username"
            f" {self.token.external_username!r}, not 'wrong'",
        ):
            serializer.save(collection=self.sid, created_by=self.scenario.user)

    def test_save_from_worker_token(self) -> None:
        worker_token = self.playground.create_worker_token()
        serializer = CollectionTokenGrantSerializer(
            data={
                "token_hash": worker_token.hash,
                "role": Collection.Roles.VIEWER,
            }
        )
        self.assertTrue(serializer.is_valid())

        with self.assertRaisesRegex(
            ValidationError, "Cannot grant roles to worker tokens"
        ):
            serializer.save(collection=self.sid, created_by=self.scenario.user)

    def test_save_from_username(self) -> None:
        serializer = CollectionTokenGrantSerializer(
            data={
                "username": "test2",
                "comment": "a comment",
                "role": Collection.Roles.VIEWER,
            }
        )
        self.assertTrue(serializer.is_valid())

        grant = serializer.save(
            collection=self.sid, created_by=self.scenario.user
        )

        self.assertGrant(grant)
        self.assertEqual(grant.token.token_type, Token.Types.EXTERNAL_USER)
        self.assertEqual(grant.token.external_username, "test2")
        self.assertEqual(grant.token.created_by, self.scenario.user)
        self.assertEqual(grant.token.comment, "a comment")
